# SPDX-FileCopyrightText: 2011 Ryan Inch
#
# SPDX-License-Identifier: GPL-2.0-or-later

import bpy

# For VARS
from . import internals

# For FUNCTIONS
from .internals import (
    update_property_group,
    get_move_selection,
    get_move_active,
)

mode_converter = {
    'EDIT_MESH': 'EDIT',
    'EDIT_CURVE': 'EDIT',
    'EDIT_SURFACE': 'EDIT',
    'EDIT_TEXT': 'EDIT',
    'EDIT_ARMATURE': 'EDIT',
    'EDIT_METABALL': 'EDIT',
    'EDIT_LATTICE': 'EDIT',
    'POSE': 'POSE',
    'SCULPT': 'SCULPT',
    'PAINT_WEIGHT': 'WEIGHT_PAINT',
    'PAINT_VERTEX': 'VERTEX_PAINT',
    'PAINT_TEXTURE': 'TEXTURE_PAINT',
    'PARTICLE': 'PARTICLE_EDIT',
    'OBJECT': 'OBJECT',
    'PAINT_GPENCIL': 'PAINT_GPENCIL',
    'EDIT_GPENCIL': 'EDIT_GPENCIL',
    'SCULPT_GPENCIL': 'SCULPT_GPENCIL',
    'WEIGHT_GPENCIL': 'WEIGHT_GPENCIL',
    'VERTEX_GPENCIL': 'VERTEX_GPENCIL',
    }


rto_path = {
    "exclude": "exclude",
    "select": "collection.hide_select",
    "hide": "hide_viewport",
    "disable": "collection.hide_viewport",
    "render": "collection.hide_render",
    "holdout": "holdout",
    "indirect": "indirect_only",
    }

set_off_on = {
    "exclude": {
        "off": True,
        "on": False
        },
    "select": {
        "off": True,
        "on": False
        },
    "hide": {
        "off": True,
        "on": False
        },
    "disable": {
        "off": True,
        "on": False
        },
    "render": {
        "off": True,
        "on": False
        },
    "holdout": {
        "off": False,
        "on": True
        },
    "indirect": {
        "off": False,
        "on": True
        }
    }

get_off_on = {
    False: {
        "exclude": "on",
        "select": "on",
        "hide": "on",
        "disable": "on",
        "render": "on",
        "holdout": "off",
        "indirect": "off",
        },

    True: {
        "exclude": "off",
        "select": "off",
        "hide": "off",
        "disable": "off",
        "render": "off",
        "holdout": "on",
        "indirect": "on",
        }
    }


def get_rto(layer_collection, rto):
    if rto in ["exclude", "hide", "holdout", "indirect"]:
        return getattr(layer_collection, rto_path[rto])

    else:
        collection = getattr(layer_collection, "collection")
        return getattr(collection, rto_path[rto].split(".")[1])


def set_rto(layer_collection, rto, value):
    if rto in ["exclude", "hide", "holdout", "indirect"]:
        setattr(layer_collection, rto_path[rto], value)

    else:
        collection = getattr(layer_collection, "collection")
        setattr(collection, rto_path[rto].split(".")[1], value)


def apply_to_children(parent, apply_function, *args, **kwargs):
    # works for both Collections & LayerCollections
    child_lists = [parent.children]

    while child_lists:
        new_child_lists = []

        for child_list in child_lists:
            for child in child_list:
                apply_function(child, *args, **kwargs)

                if child.children:
                    new_child_lists.append(child.children)

        child_lists = new_child_lists


def isolate_rto(cls, self, view_layer, rto, *, children=False):
    off = set_off_on[rto]["off"]
    on = set_off_on[rto]["on"]

    laycol_ptr = internals.layer_collections[self.name]["ptr"]
    target = internals.rto_history[rto][view_layer]["target"]
    history = internals.rto_history[rto][view_layer]["history"]

    # get active collections
    active_layer_collections = [x["ptr"] for x in internals.layer_collections.values()
                                if get_rto(x["ptr"], rto) == on]

    # check if previous state should be restored
    if cls.isolated and self.name == target:
        # restore previous state
        for x, item in enumerate(internals.layer_collections.values()):
            set_rto(item["ptr"], rto, history[x])

        # reset target and history
        del internals.rto_history[rto][view_layer]

        cls.isolated = False

    # check if all RTOs should be activated
    elif (len(active_layer_collections) == 1 and
          active_layer_collections[0].name == self.name):
        # activate all collections
        for item in internals.layer_collections.values():
            set_rto(item["ptr"], rto, on)

        # reset target and history
        del internals.rto_history[rto][view_layer]

        cls.isolated = False

    else:
        # isolate collection

        internals.rto_history[rto][view_layer]["target"] = self.name

        # reset history
        history.clear()

        # save state
        for item in internals.layer_collections.values():
            history.append(get_rto(item["ptr"], rto))

        child_states = {}
        if children:
            # get child states
            def get_child_states(layer_collection):
                child_states[layer_collection.name] = get_rto(layer_collection, rto)

            apply_to_children(laycol_ptr, get_child_states)

        # isolate collection
        for item in internals.layer_collections.values():
            if item["name"] != laycol_ptr.name:
                set_rto(item["ptr"], rto, off)

        set_rto(laycol_ptr, rto, on)

        if rto not in ["exclude", "holdout", "indirect"]:
            # activate all parents
            laycol = internals.layer_collections[self.name]
            while laycol["id"] != 0:
                set_rto(laycol["ptr"], rto, on)
                laycol = laycol["parent"]

            if children:
                # restore child states
                def restore_child_states(layer_collection):
                    set_rto(layer_collection, rto, child_states[layer_collection.name])

                apply_to_children(laycol_ptr, restore_child_states)

        else:
            if children:
                # restore child states
                def restore_child_states(layer_collection):
                    set_rto(layer_collection, rto, child_states[layer_collection.name])

                apply_to_children(laycol_ptr, restore_child_states)

            elif rto == "exclude":
                # deactivate all children
                def deactivate_all_children(layer_collection):
                    set_rto(layer_collection, rto, True)

                apply_to_children(laycol_ptr, deactivate_all_children)

        cls.isolated = True


def isolate_sel_objs_collections(view_layer, rto, caller, *, use_active=False):
    selected_objects = get_move_selection()

    if use_active:
        selected_objects.add(get_move_active(always=True))

    if not selected_objects:
        return "No selected objects"

    off = set_off_on[rto]["off"]
    on = set_off_on[rto]["on"]

    if caller == "CM":
        history = internals.rto_history[rto+"_all"][view_layer]

    elif caller == "QCD":
        history = internals.qcd_history[view_layer]


    # if not isolated, isolate collections of selected objects
    if len(history) == 0:
        keep_history = False

        # save history and isolate RTOs
        for item in internals.layer_collections.values():
            history.append(get_rto(item["ptr"], rto))
            rto_state = off

            # check if any of the selected objects are in the collection
            if not set(selected_objects).isdisjoint(item["ptr"].collection.objects):
                rto_state = on

            if history[-1] != rto_state:
                keep_history = True

            if rto == "exclude":
                set_exclude_state(item["ptr"], rto_state)

            else:
                set_rto(item["ptr"], rto, rto_state)

                # activate all parents if needed
                if rto_state == on and rto not in ["holdout", "indirect"]:
                    laycol = item["parent"]
                    while laycol["id"] != 0:
                        set_rto(laycol["ptr"], rto, on)
                        laycol = laycol["parent"]


        if not keep_history:
            history.clear()

            return "Collection already isolated"


    else:
        for x, item in enumerate(internals.layer_collections.values()):
            set_rto(item["ptr"], rto, history[x])

        # clear history
        if caller == "CM":
            del internals.rto_history[rto+"_all"][view_layer]

        elif caller == "QCD":
            del internals.qcd_history[view_layer]


def disable_sel_objs_collections(view_layer, rto, caller):
    off = set_off_on[rto]["off"]
    on = set_off_on[rto]["on"]
    selected_objects = get_move_selection()

    if caller == "CM":
        history = internals.rto_history[rto+"_all"][view_layer]

    elif caller == "QCD":
        history = internals.qcd_history[view_layer]


    if not selected_objects and not history:
        # clear history
        if caller == "CM":
            del internals.rto_history[rto+"_all"][view_layer]

        elif caller == "QCD":
            del internals.qcd_history[view_layer]

        return "No selected objects"

    # if not disabled, disable collections of selected objects
    if len(history) == 0:
        # save history and disable RTOs
        for item in internals.layer_collections.values():
            history.append(get_rto(item["ptr"], rto))

            # check if any of the selected objects are in the collection
            if not set(selected_objects).isdisjoint(item["ptr"].collection.objects):
                if rto == "exclude":
                    set_exclude_state(item["ptr"], off)

                else:
                    set_rto(item["ptr"], rto, off)


    else:
        for x, item in enumerate(internals.layer_collections.values()):
            set_rto(item["ptr"], rto, history[x])

        # clear history
        if caller == "CM":
            del internals.rto_history[rto+"_all"][view_layer]

        elif caller == "QCD":
            del internals.qcd_history[view_layer]


def toggle_children(self, view_layer, rto):
    laycol_ptr = internals.layer_collections[self.name]["ptr"]
    # clear rto history
    del internals.rto_history[rto][view_layer]
    internals.rto_history[rto+"_all"].pop(view_layer, None)

    # toggle rto state
    state = not get_rto(laycol_ptr, rto)
    set_rto(laycol_ptr, rto, state)

    def set_state(layer_collection):
        set_rto(layer_collection, rto, state)

    apply_to_children(laycol_ptr, set_state)


def activate_all_rtos(view_layer, rto):
    off = set_off_on[rto]["off"]
    on = set_off_on[rto]["on"]

    history = internals.rto_history[rto+"_all"][view_layer]

    # if not activated, activate all
    if len(history) == 0:
        keep_history = False

        for item in reversed(list(internals.layer_collections.values())):
            if get_rto(item["ptr"], rto) == off:
                keep_history = True

            history.append(get_rto(item["ptr"], rto))

            set_rto(item["ptr"], rto, on)

        if not keep_history:
            history.clear()

        history.reverse()

    else:
        for x, item in enumerate(internals.layer_collections.values()):
            set_rto(item["ptr"], rto, history[x])

        # clear rto history
        del internals.rto_history[rto+"_all"][view_layer]


def invert_rtos(view_layer, rto):
    if rto == "exclude":
        orig_values = []

        for item in internals.layer_collections.values():
            orig_values.append(get_rto(item["ptr"], rto))

        for x, item in enumerate(internals.layer_collections.values()):
            set_rto(item["ptr"], rto, not orig_values[x])

    else:
        for item in internals.layer_collections.values():
            set_rto(item["ptr"], rto, not get_rto(item["ptr"], rto))

    # clear rto history
    internals.rto_history[rto].pop(view_layer, None)


def copy_rtos(view_layer, rto):
    if not internals.copy_buffer["RTO"]:
        # copy
        internals.copy_buffer["RTO"] = rto
        for laycol in internals.layer_collections.values():
            internals.copy_buffer["values"].append(get_off_on[
                                            get_rto(laycol["ptr"], rto)
                                            ][
                                            rto
                                            ]
                                        )

    else:
        # paste
        for x, laycol in enumerate(internals.layer_collections.values()):
            set_rto(laycol["ptr"],
                    rto,
                    set_off_on[rto][
                        internals.copy_buffer["values"][x]
                        ]
                    )

        # clear rto history
        internals.rto_history[rto].pop(view_layer, None)
        del internals.rto_history[rto+"_all"][view_layer]

        # clear copy buffer
        internals.copy_buffer["RTO"] = ""
        internals.copy_buffer["values"].clear()


def swap_rtos(view_layer, rto):
    if not internals.swap_buffer["A"]["values"]:
        # get A
        internals.swap_buffer["A"]["RTO"] = rto
        for laycol in internals.layer_collections.values():
            internals.swap_buffer["A"]["values"].append(get_off_on[
                                                get_rto(laycol["ptr"], rto)
                                                ][
                                                rto
                                                ]
                                            )

    else:
        # get B
        internals.swap_buffer["B"]["RTO"] = rto
        for laycol in internals.layer_collections.values():
            internals.swap_buffer["B"]["values"].append(get_off_on[
                                                get_rto(laycol["ptr"], rto)
                                                ][
                                                rto
                                                ]
                                            )

        # swap A with B
        for x, laycol in enumerate(internals.layer_collections.values()):
            set_rto(laycol["ptr"], internals.swap_buffer["A"]["RTO"],
                    set_off_on[
                        internals.swap_buffer["A"]["RTO"]
                        ][
                        internals.swap_buffer["B"]["values"][x]
                        ]
                    )

            set_rto(laycol["ptr"], internals.swap_buffer["B"]["RTO"],
                    set_off_on[
                        internals.swap_buffer["B"]["RTO"]
                        ][
                        internals.swap_buffer["A"]["values"][x]
                        ]
                    )


        # clear rto history
        swap_a = internals.swap_buffer["A"]["RTO"]
        swap_b = internals.swap_buffer["B"]["RTO"]

        internals.rto_history[swap_a].pop(view_layer, None)
        internals.rto_history[swap_a+"_all"].pop(view_layer, None)
        internals.rto_history[swap_b].pop(view_layer, None)
        internals.rto_history[swap_b+"_all"].pop(view_layer, None)

        # clear swap buffer
        internals.swap_buffer["A"]["RTO"] = ""
        internals.swap_buffer["A"]["values"].clear()
        internals.swap_buffer["B"]["RTO"] = ""
        internals.swap_buffer["B"]["values"].clear()


def clear_copy(rto):
    if internals.copy_buffer["RTO"] == rto:
        internals.copy_buffer["RTO"] = ""
        internals.copy_buffer["values"].clear()


def clear_swap(rto):
    if internals.swap_buffer["A"]["RTO"] == rto:
        internals.swap_buffer["A"]["RTO"] = ""
        internals.swap_buffer["A"]["values"].clear()
        internals.swap_buffer["B"]["RTO"] = ""
        internals.swap_buffer["B"]["values"].clear()


def link_child_collections_to_parent(laycol, collection, parent_collection):
    # store view layer RTOs for all children of the to be deleted collection
    child_states = {}
    def get_child_states(layer_collection):
        child_states[layer_collection.name] = (layer_collection.exclude,
                                               layer_collection.hide_viewport,
                                               layer_collection.holdout,
                                               layer_collection.indirect_only)

    apply_to_children(laycol["ptr"], get_child_states)

    # link any subcollections of the to be deleted collection to it's parent
    for subcollection in collection.children:
        if not subcollection.name in parent_collection.children:
            parent_collection.children.link(subcollection)

    # apply the stored view layer RTOs to the newly linked collections and their
    # children
    def restore_child_states(layer_collection):
        state = child_states.get(layer_collection.name)

        if state:
            layer_collection.exclude = state[0]
            layer_collection.hide_viewport = state[1]
            layer_collection.holdout = state[2]
            layer_collection.indirect_only = state[3]

    apply_to_children(laycol["parent"]["ptr"], restore_child_states)


def remove_collection(laycol, collection, context):
    # get selected row
    cm = context.scene.collection_manager
    selected_row_name = cm.cm_list_collection[cm.cm_list_index].name

    # delete collection
    bpy.data.collections.remove(collection)

    # update references
    internals.expanded.discard(laycol["name"])

    if internals.expand_history["target"] == laycol["name"]:
        internals.expand_history["target"] = ""

    if laycol["name"] in internals.expand_history["history"]:
        internals.expand_history["history"].remove(laycol["name"])

    if internals.qcd_slots.contains(name=laycol["name"]):
        internals.qcd_slots.del_slot(name=laycol["name"])

    if laycol["name"] in internals.qcd_slots.overrides:
        internals.qcd_slots.overrides.remove(laycol["name"])

    # reset history
    for rto in internals.rto_history.values():
        rto.clear()

    # update tree view
    update_property_group(context)

    # update selected row
    laycol = internals.layer_collections.get(selected_row_name, None)
    if laycol:
        cm.cm_list_index = laycol["row_index"]

    elif len(cm.cm_list_collection) <= cm.cm_list_index:
        cm.cm_list_index =  len(cm.cm_list_collection) - 1

        if cm.cm_list_index > -1:
            name = cm.cm_list_collection[cm.cm_list_index].name
            laycol = internals.layer_collections[name]
            while not laycol["visible"]:
                laycol = laycol["parent"]

            cm.cm_list_index = laycol["row_index"]


def select_collection_objects(is_master_collection, collection_name, replace, nested, selection_state=None):
    if bpy.context.mode != 'OBJECT':
        return

    if is_master_collection:
        target_collection = bpy.context.view_layer.layer_collection.collection

    else:
        laycol = internals.layer_collections[collection_name]
        target_collection = laycol["ptr"].collection

    if replace:
        bpy.ops.object.select_all(action='DESELECT')

    if selection_state == None:
        selection_state = get_move_selection().isdisjoint(target_collection.objects)

    def select_objects(collection, selection_state):
        for obj in collection.objects:
            try:
                obj.select_set(selection_state)
            except RuntimeError:
                pass

    select_objects(target_collection, selection_state)

    if nested:
        apply_to_children(target_collection, select_objects, selection_state)

def set_exclude_state(target_layer_collection, state):
    # get current child exclusion state
    child_exclusion = []

    def get_child_exclusion(layer_collection):
        child_exclusion.append([layer_collection, layer_collection.exclude])

    apply_to_children(target_layer_collection, get_child_exclusion)


    # set exclusion
    target_layer_collection.exclude = state


    # set correct state for all children
    for laycol in child_exclusion:
        laycol[0].exclude = laycol[1]



# SPDX-FileCopyrightText: 2017-2022 Blender Foundation
#
# SPDX-License-Identifier: GPL-2.0-or-later

__author__ = "Alexander Milovsky, Nutti <nutti.metro@gmail.com>"
__status__ = "production"
__version__ = "6.6"
__date__ = "22 Apr 2022"

from math import sin, cos, pi

import bpy
import bmesh
from bpy.props import (
    FloatProperty,
    FloatVectorProperty,
    BoolProperty,
    EnumProperty
)
from mathutils import Vector

from .. import common
from ..utils.bl_class_registry import BlClassRegistry
from ..utils.property_class_registry import PropertyClassRegistry
from ..utils import compatibility as compat


def _is_valid_context(context):
    # only 'VIEW_3D' space is allowed to execute
    if not common.is_valid_space(context, ['VIEW_3D']):
        return False

    objs = common.get_uv_editable_objects(context)
    if not objs:
        return False

    # only edit mode is allowed to execute
    if context.object.mode != 'EDIT':
        return False

    return True


def _get_uv_layer(ops_obj, bm, assign_uvmap):
    # get UV layer
    if not bm.loops.layers.uv:
        if assign_uvmap:
            bm.loops.layers.uv.new()
        else:
            ops_obj.report({'WARNING'},
                           "Object must have more than one UV map")
            return None
    uv_layer = bm.loops.layers.uv.verify()

    return uv_layer


def _apply_box_map(bm, uv_layer, size, offset, rotation,
                   tex_aspect, force_axis, force_axis_tex_aspect_correction,
                   force_axis_rotation):
    scale = 1.0 / size

    sx = 1.0 * scale
    sy = 1.0 * scale
    sz = 1.0 * scale
    ofx = offset[0]
    ofy = offset[1]
    ofz = offset[2]
    rx = rotation[0] * pi / 180.0
    ry = rotation[1] * pi / 180.0
    rz = rotation[2] * pi / 180.0

    farx = force_axis_rotation[0] * pi / 180.0
    fary = force_axis_rotation[1] * pi / 180.0
    farz = force_axis_rotation[2] * pi / 180.0

    sel_faces = [f for f in bm.faces if f.select]

    # update UV coordinate
    for f in sel_faces:
        n = f.normal
        for l in f.loops:
            co = l.vert.co
            x = co.x * sx
            y = co.y * sy
            z = co.z * sz
            aspect = tex_aspect

            transformed = False
            if force_axis == 'X':
                # Use Y-plane
                if abs(n[1]) < abs(n[0]) and abs(n[1]) >= abs(n[2]):
                    aspect *= force_axis_tex_aspect_correction
                    if n[1] >= 0.0:
                        u = -(x - ofx) * cos(fary) + (z - ofz) * sin(fary)
                        v = (x * aspect - ofx) * sin(fary) + \
                            (z * aspect - ofz) * cos(fary)
                    else:
                        u = (x - ofx) * cos(fary) + (z - ofz) * sin(fary)
                        v = -(x * aspect - ofx) * sin(fary) + \
                            (z * aspect - ofz) * cos(fary)
                    transformed = True
                # Use Z-plane
                elif abs(n[2]) < abs(n[0]) and abs(n[2]) >= abs(n[1]):
                    aspect *= force_axis_tex_aspect_correction
                    if n[2] >= 0.0:
                        u = (x - ofx) * cos(farz) + (y - ofy) * sin(farz)
                        v = -(x * aspect - ofx) * sin(farz) + \
                            (y * aspect - ofy) * cos(farz)
                    else:
                        u = -(x - ofx) * cos(farz) - (y + ofy) * sin(farz)
                        v = -(x * aspect + ofx) * sin(farz) + \
                            (y * aspect - ofy) * cos(farz)
                    transformed = True
            elif force_axis == 'Y':
                # Use X-plane
                if abs(n[0]) < abs(n[1]) and abs(n[0]) >= abs(n[2]):
                    aspect *= force_axis_tex_aspect_correction
                    if n[0] >= 0.0:
                        u = (y - ofy) * cos(farx) + (z - ofz) * sin(farx)
                        v = -(y * aspect - ofy) * sin(farx) + \
                            (z * aspect - ofz) * cos(farx)
                    else:
                        u = -(y - ofy) * cos(farx) + (z - ofz) * sin(farx)
                        v = (y * aspect - ofy) * sin(farx) + \
                            (z * aspect - ofz) * cos(farx)
                    transformed = True
                # Use Z-plane
                elif abs(n[2]) >= abs(n[0]) and abs(n[2]) < abs(n[1]):
                    aspect *= force_axis_tex_aspect_correction
                    if n[2] >= 0.0:
                        u = (x - ofx) * cos(farz) + (y - ofy) * sin(farz)
                        v = -(x * aspect - ofx) * sin(farz) + \
                            (y * aspect - ofy) * cos(farz)
                    else:
                        u = -(x - ofx) * cos(farz) - (y + ofy) * sin(farz)
                        v = -(x * aspect + ofx) * sin(farz) + \
                            (y * aspect - ofy) * cos(farz)
                    transformed = True
            elif force_axis == 'Z':
                # Use X-plane
                if abs(n[0]) >= abs(n[1]) and abs(n[0]) < abs(n[2]):
                    aspect *= force_axis_tex_aspect_correction
                    if n[0] >= 0.0:
                        u = (y - ofy) * cos(farx) + (z - ofz) * sin(farx)
                        v = -(y * aspect - ofy) * sin(farx) + \
                            (z * aspect - ofz) * cos(farx)
                    else:
                        u = -(y - ofy) * cos(farx) + (z - ofz) * sin(farx)
                        v = (y * aspect - ofy) * sin(farx) + \
                            (z * aspect - ofz) * cos(farx)
                    transformed = True
                # Use Y-plane
                elif abs(n[1]) >= abs(n[0]) and abs(n[1]) < abs(n[2]):
                    aspect *= force_axis_tex_aspect_correction
                    if n[1] >= 0.0:
                        u = -(x - ofx) * cos(fary) + (z - ofz) * sin(fary)
                        v = (x * aspect - ofx) * sin(fary) + \
                            (z * aspect - ofz) * cos(fary)
                    else:
                        u = (x - ofx) * cos(fary) + (z - ofz) * sin(fary)
                        v = -(x * aspect - ofx) * sin(fary) + \
                            (z * aspect - ofz) * cos(fary)
                    transformed = True

            if not transformed:
                # X-plane
                if abs(n[0]) >= abs(n[1]) and abs(n[0]) >= abs(n[2]):
                    if n[0] >= 0.0:
                        u = (y - ofy) * cos(rx) + (z - ofz) * sin(rx)
                        v = -(y * aspect - ofy) * sin(rx) + \
                            (z * aspect - ofz) * cos(rx)
                    else:
                        u = -(y - ofy) * cos(rx) + (z - ofz) * sin(rx)
                        v = (y * aspect - ofy) * sin(rx) + \
                            (z * aspect - ofz) * cos(rx)
                # Y-plane
                elif abs(n[1]) >= abs(n[0]) and abs(n[1]) >= abs(n[2]):
                    if n[1] >= 0.0:
                        u = -(x - ofx) * cos(ry) + (z - ofz) * sin(ry)
                        v = (x * aspect - ofx) * sin(ry) + \
                            (z * aspect - ofz) * cos(ry)
                    else:
                        u = (x - ofx) * cos(ry) + (z - ofz) * sin(ry)
                        v = -(x * aspect - ofx) * sin(ry) + \
                            (z * aspect - ofz) * cos(ry)
                # Z-plane
                elif abs(n[2]) >= abs(n[0]) and abs(n[2]) >= abs(n[1]):
                    if n[2] >= 0.0:
                        u = (x - ofx) * cos(rz) + (y - ofy) * sin(rz)
                        v = -(x * aspect - ofx) * sin(rz) + \
                            (y * aspect - ofy) * cos(rz)
                    else:
                        u = -(x - ofx) * cos(rz) - (y + ofy) * sin(rz)
                        v = -(x * aspect + ofx) * sin(rz) + \
                            (y * aspect - ofy) * cos(rz)

            l[uv_layer].uv = Vector((u, v))


def _apply_planer_map(bm, uv_layer, size, offset, rotation, tex_aspect):
    scale = 1.0 / size

    sx = 1.0 * scale
    sy = 1.0 * scale
    ofx = offset[0]
    ofy = offset[1]
    rz = rotation * pi / 180.0
    aspect = tex_aspect

    sel_faces = [f for f in bm.faces if f.select]

    # calculate average of normal
    n_ave = Vector((0.0, 0.0, 0.0))
    for f in sel_faces:
        n_ave = n_ave + f.normal
    q = n_ave.rotation_difference(Vector((0.0, 0.0, 1.0)))

    # update UV coordinate
    for f in sel_faces:
        for l in f.loops:
            co = compat.matmul(q, l.vert.co)
            x = co.x * sx
            y = co.y * sy

            u = x * cos(rz) - y * sin(rz) + ofx
            v = -x * aspect * sin(rz) - y * aspect * cos(rz) + ofy

            l[uv_layer].uv = Vector((u, v))


@PropertyClassRegistry()
class _Properties:
    idname = "uvw"

    @classmethod
    def init_props(cls, scene):
        scene.muv_uvw_enabled = BoolProperty(
            name="UVW Enabled",
            description="UVW is enabled",
            default=False
        )
        scene.muv_uvw_assign_uvmap = BoolProperty(
            name="Assign UVMap",
            description="Assign UVMap when no UVmaps are available",
            default=True
        )

    @classmethod
    def del_props(cls, scene):
        del scene.muv_uvw_enabled
        del scene.muv_uvw_assign_uvmap


@BlClassRegistry()
@compat.make_annotations
class MUV_OT_UVW_BoxMap(bpy.types.Operator):
    bl_idname = "uv.muv_uvw_box_map"
    bl_label = "Box Map"
    bl_options = {'REGISTER', 'UNDO'}

    size = FloatProperty(
        name="Size",
        default=1.0,
        precision=4
    )
    rotation = FloatVectorProperty(
        name="Rotation",
        size=3,
        default=(0.0, 0.0, 0.0),
        subtype='XYZ'
    )
    offset = FloatVectorProperty(
        name="Offset",
        size=3,
        default=(0.0, 0.0, 0.0),
        subtype='XYZ'
    )
    tex_aspect = FloatProperty(
        name="Texture Aspect",
        default=1.0,
        precision=4
    )
    assign_uvmap = BoolProperty(
        name="Assign UVMap",
        description="Assign UVMap when no UVmaps are available",
        default=True
    )
    force_axis = EnumProperty(
        name="Force Axis",
        description="Axis to force the mapping",
        items=[
            ('NONE', "None", "None"),
            ('X', "X", "Axis X"),
            ('Y', "Y", "Axis Y"),
            ('Z', "Z", "Axis Z")
        ],
        default='NONE'
    )
    force_axis_tex_aspect_correction = FloatProperty(
        name="Texture Aspect Correction (Force Axis)",
        description="Texture Aspect correction for the faces mapped forcibly",
        default=3.14,
        precision=4
    )
    force_axis_rotation = FloatVectorProperty(
        name="Rotation (Force Axis)",
        description="Rotation for the faces mapped forcibly",
        size=3,
        default=(0.0, 0.0, 0.0),
        subtype='XYZ'
    )

    @classmethod
    def poll(cls, context):
        # we can not get area/space/region from console
        if common.is_console_mode():
            return True
        return _is_valid_context(context)

    def draw(self, _):
        layout = self.layout

        col = layout.column()
        row = col.row()
        row.label(text="Size:")
        row.prop(self, "size", text="")

        layout.label(text="Rotation:")
        layout.row().prop(self, "rotation", text="")

        layout.label(text="Offset:")
        layout.row().prop(self, "offset", text="")

        col = layout.column()
        row = col.row()
        row.label(text="Texture Aspect:")
        row.prop(self, "tex_aspect", text="")

        layout.prop(self, "assign_uvmap")

        layout.separator(factor=2.0)

        layout.prop(self, "force_axis")
        if self.force_axis != 'NONE':
            col = layout.column()
            row = col.row()
            row.label(text="Texture Aspect Correction (Force Axis)")
            row.prop(self, "force_axis_tex_aspect_correction", text="")

            layout.label(text="Rotation (Force Axis)")
            layout.row().prop(self, "force_axis_rotation", text="")

    def execute(self, context):
        objs = common.get_uv_editable_objects(context)

        for obj in objs:
            bm = bmesh.from_edit_mesh(obj.data)
            if common.check_version(2, 73, 0) >= 0:
                bm.faces.ensure_lookup_table()

            # get UV layer
            uv_layer = _get_uv_layer(self, bm, self.assign_uvmap)
            if not uv_layer:
                return {'CANCELLED'}

            _apply_box_map(bm, uv_layer, self.size, self.offset, self.rotation,
                           self.tex_aspect, self.force_axis,
                           self.force_axis_tex_aspect_correction,
                           self.force_axis_rotation)
            bmesh.update_edit_mesh(obj.data)

        return {'FINISHED'}


@BlClassRegistry()
@compat.make_annotations
class MUV_OT_UVW_BestPlanerMap(bpy.types.Operator):
    bl_idname = "uv.muv_uvw_best_planer_map"
    bl_label = "Best Planer Map"
    bl_options = {'REGISTER', 'UNDO'}

    size = FloatProperty(
        name="Size",
        default=1.0,
        precision=4
    )
    rotation = FloatProperty(
        name="Rotation",
        default=0.0
    )
    offset = FloatVectorProperty(
        name="Offset",
        size=2,
        default=(0.0, 0.0)
    )
    tex_aspect = FloatProperty(
        name="Texture Aspect",
        default=1.0,
        precision=4
    )
    assign_uvmap = BoolProperty(
        name="Assign UVMap",
        description="Assign UVMap when no UVmaps are available",
        default=True
    )

    @classmethod
    def poll(cls, context):
        # we can not get area/space/region from console
        if common.is_console_mode():
            return True
        return _is_valid_context(context)

    def execute(self, context):
        objs = common.get_uv_editable_objects(context)

        for obj in objs:
            bm = bmesh.from_edit_mesh(obj.data)
            if common.check_version(2, 73, 0) >= 0:
                bm.faces.ensure_lookup_table()

            # get UV layer
            uv_layer = _get_uv_layer(self, bm, self.assign_uvmap)
            if not uv_layer:
                return {'CANCELLED'}

            _apply_planer_map(bm, uv_layer, self.size, self.offset,
                              self.rotation, self.tex_aspect)

            bmesh.update_edit_mesh(obj.data)

        return {'FINISHED'}



# SPDX-FileCopyrightText: 2019-2022 Blender Foundation
#
# SPDX-License-Identifier: GPL-2.0-or-later

import bpy
from bpy.types import Operator, Panel
from bpy.props import (StringProperty,
                       EnumProperty,
                       FloatProperty,
                       BoolProperty)

from io_mesh_atomic.utility_panel import choose_objects
from io_mesh_atomic.utility_panel import custom_datafile
from io_mesh_atomic.utility_panel import custom_datafile_change_atom_props
from io_mesh_atomic.utility_panel import separate_atoms
from io_mesh_atomic.utility_panel import distance


# -----------------------------------------------------------------------------
#                                                                           GUI

# The panel.
class PANEL_PT_prepare(Panel):
    bl_label       = "Atomic Blender Utilities"
    bl_space_type  = "VIEW_3D"
    bl_region_type = "UI"
    bl_options     = {'DEFAULT_CLOSED'}
    bl_category = "Create"
    bl_idname = "ATOMIC_PT_utilities"


    # This thing here just guarantees that the panel is NOT opened when the
    # check box in the addon preferences is not activated! See __init__.py
    @classmethod
    def poll(cls, context):
        pref = context.preferences
        return pref.addons[__package__].preferences.bool_utility

    def draw(self, context):
        layout = self.layout
        scn    = context.scene.atom_blend

        box = layout.box()
        col = box.column(align=True)
        col.label(text="Custom data file")
        col.prop(scn, "datafile")
        col.operator("atom_blend.datafile_apply")

        box = layout.box()
        col = box.column(align=True)
        col.label(text="Measure distances")
        col.operator("atom_blend.button_distance")
        col.prop(scn, "distance")

        box = layout.box()
        col = box.column(align=True)
        col.label(text="Change atom size")
        col.label(text="1. Type of radii")
        col.prop(scn, "radius_type")
        col2 = col.column()
        col2.active = (scn.radius_type == '3')
        col2.prop(scn, "radius_type_ionic")
        col = box.column(align=True)
        col.label(text="2. Radii in pm")
        col.prop(scn, "radius_pm_name")
        col.prop(scn, "radius_pm")
        col = box.column(align=True)
        col.label(text="3. Radii by scale")
        col.prop(scn, "radius_all")
        row = col.row()
        row.operator("atom_blend.radius_all_smaller")
        row.operator("atom_blend.radius_all_bigger")

        box = layout.box()
        col = box.column(align=True)
        col.label(text="Change stick size")
        col.prop(scn, "sticks_all")
        row = col.row()
        row.operator("atom_blend.sticks_all_smaller")
        row.operator("atom_blend.sticks_all_bigger")

        box = layout.box()
        col = box.column(align=True)
        col.label(text="Change atom shape")
        col2 = col.column()
        col2.active = (scn.replace_objs_special == '0')
        col2.prop(scn, "replace_objs")
        col2.prop(scn, "replace_objs_material")
        col.prop(scn, "replace_objs_special")
        col.operator("atom_blend.replace_atom")
        col.label(text="Default values")
        col.operator("atom_blend.default_atoms")

        box = layout.box()
        col = box.column(align=True)
        col.label(text="Separate atoms")
        col3 = col.column()
        col3.active = (bpy.context.mode == 'EDIT_MESH')
        col3.operator("atom_blend.separate_atom")


# The properties of buttons etc. in the panel.
class PanelProperties(bpy.types.PropertyGroup):

    def Callback_radius_type(self, context):
        scn = bpy.context.scene.atom_blend
        choose_objects("ATOM_RADIUS_TYPE",
                       None,
                       None,
                       scn.radius_type,
                       scn.radius_type_ionic,
                       None)
    def Callback_radius_pm(self, context):
        scn = bpy.context.scene.atom_blend
        choose_objects("ATOM_RADIUS_PM",
                       None,
                       [scn.radius_pm_name,
                       scn.radius_pm],
                       None,
                       None,
                       None)

    datafile: StringProperty(
        name = "", description="Path to your custom data file",
        maxlen = 256, default = "", subtype='FILE_PATH')
    XYZ_file: StringProperty(
        name = "Path to file", default="",
        description = "Path of the XYZ file")
    number_atoms: StringProperty(name="",
        default="Number", description = "This output shows "
        "the number of atoms which have been loaded")
    distance: StringProperty(
        name="", default="Distance (A)",
        description="Distance of 2 objects in Angstrom")
    replace_objs: EnumProperty(
        name="Shape",
        description="Choose a different atom shape",
        items=(('0',"Unchanged", "Do not change the shape"),
               ('1a',"Sphere (Mesh)", "Replace with a sphere (Mesh)"),
               ('1b',"Sphere (NURBS)", "Replace with a sphere (NURBS)"),
               ('2',"Cube", "Replace with a cube"),
               ('3',"Plane", "Replace with a plane"),
               ('4a',"Circle (Mesh)", "Replace with a circle (Mesh)"),
               ('4b',"Circle (NURBS)", "Replace with a circle (NURBS)"),
               ('5a',"Icosphere 1", "Replace with a icosphere, subd=1"),
               ('5b',"Icosphere 2", "Replace with a icosphere, subd=2"),
               ('5c',"Icosphere 3", "Replace with a icosphere, subd=3"),
               ('5d',"Icosphere 4", "Replace with a icosphere, subd=4"),
               ('5e',"Icosphere 5", "Replace with a icosphere, subd=5"),
               ('6a',"Cylinder (Mesh)", "Replace with a cylinder (Mesh)"),
               ('6b',"Cylinder (NURBS)", "Replace with a cylinder (NURBS)"),
               ('7',"Cone", "Replace with a cone"),
               ('8a',"Torus (Mesh)", "Replace with a torus (Mesh)"),
               ('8b',"Torus (NURBS)", "Replace with a torus (NURBS)")),
               default='0',)
    replace_objs_material: EnumProperty(
        name="Material",
        description="Choose a different material",
        items=(('0',"Unchanged", "Leave the material unchanged"),
               ('1',"Normal", "Use normal material (no transparency and reflection)"),
               ('2',"Transparent", "Use transparent material"),
               ('3',"Reflecting", "Use reflecting material"),
               ('4',"Transparent + reflecting", "Use transparent and reflecting material")),
               default='0',)
    replace_objs_special: EnumProperty(
        name="Special",
        description="Choose a special atom shape",
        items=(('0',"None", "Use no special shape."),
               ('1',"F2+ center", "Replace with a F2+ center"),
               ('2',"F+ center", "Replace with a F+ center"),
               ('3',"F0 center", "Replace with a F0 center")),
               default='0',)
    radius_type: EnumProperty(
        name="Type",
        description="Which type of atom radii?",
        items=(('0',"predefined", "Use pre-defined radii"),
               ('1',"atomic", "Use atomic radii"),
               ('2',"van der Waals","Use van der Waals radii"),
               ('3',"ionic radii", "Use ionic radii")),
               default='0',update=Callback_radius_type)
    radius_type_ionic: EnumProperty(
        name="Charge",
        description="Charge state of the ions if existing",
        items=(('0',"-4", "Charge state -4"),
               ('1',"-3", "Charge state -3"),
               ('2',"-2", "Charge state -2"),
               ('3',"-1", "Charge state -1"),
               ('4'," 0", "Charge state  0: nothing is done"),
               ('5',"+1", "Charge state +1"),
               ('6',"+2", "Charge state +2"),
               ('7',"+3", "Charge state +3"),
               ('8',"+4", "Charge state +4"),
               ('9',"+5", "Charge state +5"),
               ('10',"+6", "Charge state +6"),
               ('11',"+7", "Charge state +7")),
               default='4',update=Callback_radius_type)
    radius_pm_name: StringProperty(
        name="", default="Atom name",
        description="Put in the name of the atom (e.g. Hydrogen)")
    radius_pm: FloatProperty(
        name="", default=100.0, min=0.0,
        description="Put in the radius of the atom (in pm)",
        update=Callback_radius_pm)
    radius_all: FloatProperty(
        name="Scale", default = 1.05, min=1.0, max=5.0,
        description="Put in the scale factor")
    sticks_all: FloatProperty(
        name="Scale", default = 1.05, min=1.0, max=5.0,
        description="Put in the scale factor")


# Button loading a custom data file
class DatafileApply(Operator):
    bl_idname = "atom_blend.datafile_apply"
    bl_label = "Apply"
    bl_description = "Use color and radii values stored in the custom file"

    def execute(self, context):
        scn = bpy.context.scene.atom_blend

        if scn.datafile == "":
            return {'FINISHED'}

        custom_datafile(scn.datafile)
        custom_datafile_change_atom_props()

        return {'FINISHED'}


# Button for separating single atoms from a dupliverts structure
class DefaultAtom(Operator):
    bl_idname = "atom_blend.default_atoms"
    bl_label = "Default"
    bl_description = ("Use default shapes and colors for atoms")

    # Are we in the OBJECT mode?
    @classmethod
    def poll(self, context):
        if bpy.context.mode == 'OBJECT':
            return True
        else:
            return False

    def execute(self, context):
        scn = bpy.context.scene.atom_blend
        choose_objects("ATOM_DEFAULT_OBJ",
                       None,
                       None,
                       None,
                       None,
                       None)
        return {'FINISHED'}


# Button for separating single atoms from a dupliverts structure
class ReplaceAtom(Operator):
    bl_idname = "atom_blend.replace_atom"
    bl_label = "Replace"
    bl_description = ("Replace selected atoms with atoms of different shape")

    # Are we in the OBJECT mode?
    @classmethod
    def poll(self, context):
        if bpy.context.mode == 'OBJECT':
            return True
        else:
            return False

    def execute(self, context):
        scn = bpy.context.scene.atom_blend
        choose_objects("ATOM_REPLACE_OBJ",
                       None,
                       None,
                       None,
                       None,
                       None)
        return {'FINISHED'}


# Button for separating single atoms from a dupliverts structure
class SeparateAtom(Operator):
    bl_idname = "atom_blend.separate_atom"
    bl_label = "Separate"
    bl_description = ("Separate selected atoms in a dupliverts structure. "
                      "You have to be in the 'Edit Mode'")

    # Are we in the EDIT mode?
    @classmethod
    def poll(self, context):
        if bpy.context.mode == 'EDIT_MESH':
            return True
        else:
            return False

    def execute(self, context):
        scn = bpy.context.scene.atom_blend

        separate_atoms(scn)

        return {'FINISHED'}


# Button for measuring the distance of active objects
class DistanceButton(Operator):
    bl_idname = "atom_blend.button_distance"
    bl_label = "Measure ..."
    bl_description = "Measure the distance between two atoms (objects)"

    def execute(self, context):
        scn  = bpy.context.scene.atom_blend
        dist = distance()

        # Put the distance into the string of the output field.
        scn.distance = dist
        return {'FINISHED'}


# Button for increasing the radii of all selected atoms
class RadiusAllBiggerButton(Operator):
    bl_idname = "atom_blend.radius_all_bigger"
    bl_label = "Bigger ..."
    bl_description = "Increase the radii of selected atoms"

    def execute(self, context):
        scn = bpy.context.scene.atom_blend
        choose_objects("ATOM_RADIUS_ALL",
                       scn.radius_all,
                       None,
                       None,
                       None,
                       None)
        return {'FINISHED'}


# Button for decreasing the radii of all selected atoms
class RadiusAllSmallerButton(Operator):
    bl_idname = "atom_blend.radius_all_smaller"
    bl_label = "Smaller ..."
    bl_description = "Decrease the radii of selected atoms"

    def execute(self, context):
        scn = bpy.context.scene.atom_blend
        choose_objects("ATOM_RADIUS_ALL",
                       1.0/scn.radius_all,
                       None,
                       None,
                       None,
                       None)
        return {'FINISHED'}


# Button for increasing the radii of all selected sticks
class SticksAllBiggerButton(Operator):
    bl_idname = "atom_blend.sticks_all_bigger"
    bl_label = "Bigger ..."
    bl_description = "Increase the radii of selected sticks"

    def execute(self, context):
        scn = bpy.context.scene.atom_blend
        choose_objects("STICKS_RADIUS_ALL",
                       None,
                       None,
                       None,
                       None,
                       scn.sticks_all)
        return {'FINISHED'}


# Button for decreasing the radii of all selected sticks
class SticksAllSmallerButton(Operator):
    bl_idname = "atom_blend.sticks_all_smaller"
    bl_label = "Smaller ..."
    bl_description = "Decrease the radii of selected sticks"

    def execute(self, context):
        scn = bpy.context.scene.atom_blend
        choose_objects("STICKS_RADIUS_ALL",
                       None,
                       None,
                       None,
                       None,
                       1.0/scn.sticks_all)
        return {'FINISHED'}



# SPDX-FileCopyrightText: 2016-2020 by Nathan Lovato, Daniel Oakey, Razvan Radulescu, and contributors
#
# SPDX-License-Identifier: GPL-3.0-or-later


import bpy

from .utils.doc import doc_name, doc_idname, doc_brief, doc_description


class POWER_SEQUENCER_OT_save_direct(bpy.types.Operator):
    """
    Saves current file without prompting for confirmation. Overrides Blender default
    """

    doc = {
        "name": doc_name(__qualname__),
        "demo": "",
        "description": doc_description(__doc__),
        "shortcuts": [({"type": "S", "value": "PRESS", "ctrl": True}, {}, "Direct Save")],
        "keymap": "Sequencer",
    }
    bl_idname = doc_idname(__qualname__)
    bl_label = doc["name"]
    bl_description = doc_brief(doc["description"])
    bl_options = {"REGISTER", "UNDO"}

    @classmethod
    def poll(cls, context):
        return context.scene

    def execute(self, context):
        if bpy.data.is_saved:
            bpy.ops.wm.save_mainfile()
        else:
            # FIXME: what does this override do?
            context_override = {"dict": "override"}
            with context.temp_override(**context_override):
                bpy.ops.wm.save_as_mainfile('INVOKE_DEFAULT')
        self.report({"INFO"}, "File saved")
        return {"FINISHED"}



# SPDX-FileCopyrightText: 2012 Manfred Moitzi (mozman)
#
# SPDX-License-Identifier: MIT

# Purpose: handle default chunk
# Created: 21.07.2012, taken from my ezdxf project
from __future__ import unicode_literals
__author__ = "mozman <mozman@gmx.at>"

from .tags import Tags, DXFTag


class DefaultChunk(object):
    def __init__(self, tags):
        assert isinstance(tags, Tags)
        self.tags = tags

    @staticmethod
    def from_tags(tags, drawing):
        return DefaultChunk(tags)

    @property
    def name(self):
        return self.tags[1].value.lower()


def iterchunks(tagreader, stoptag='EOF', endofchunk='ENDSEC'):
    while True:
        tag = next(tagreader)
        if tag == DXFTag(0, stoptag):
            return

        tags = Tags([tag])
        append = tags.append
        end_tag = DXFTag(0, endofchunk)
        while tag != end_tag:
            tag = next(tagreader)
            append(tag)
        yield tags



# SPDX-FileCopyrightText: 2011-2023 Blender Foundation
#
# SPDX-License-Identifier: GPL-2.0-or-later

"""
Even though this is in a package this can run as a stand alone scripts.

# --- example usage
blender --python release/scripts/addons/system_demo_mode/demo_mode.py

looks for demo.py textblock or file in the same path as the blend:
# --- example
config = [
    dict(anim_cycles=1, anim_render=False, anim_screen_switch=0.0, anim_time_max=10.0, anim_time_min=4.0, mode='AUTO', display_render=4.0, file='/l/19534_simplest_mesh_2.blend'),
    dict(anim_cycles=1, anim_render=False, anim_screen_switch=0.0, anim_time_max=10.0, anim_time_min=4.0, mode='AUTO', display_render=4.0, file='/l/252_pivotConstraint_01.blend'),
    ]
# ---
/data/src/blender/lib/tests/rendering/
"""

import bpy
import time
import os

from . import DEMO_CFG

# populate from script
global_config_files = []

global_config = dict(
    anim_cycles=1,
    anim_render=False,
    anim_screen_switch=0.0,
    anim_time_max=60.0,
    anim_time_min=4.0,
    mode='AUTO',
    display_render=4.0,
)

# switch to the next file in 2 sec.
global_config_fallback = dict(
    anim_cycles=1,
    anim_render=False,
    anim_screen_switch=0.0,
    anim_time_max=60.0,
    anim_time_min=4.0,
    mode='AUTO',
    display_render=4.0,
)

global_state = {
    "init_time": 0.0,
    "last_switch": 0.0,
    "reset_anim": False,
    "anim_cycles": 0,  # count how many times we played the anim
    "last_frame": -1,
    "is_render": False,
    "render_time": "",  # time render was finished.
    "timer": None,
    "basedir": "",  # demo.py is stored here
    "demo_index": 0,
    "exit": False,
}


# -----------------------------------------------------------------------------
# render handler - maintain "is_render"

def handle_render_clear():
    for ls in (bpy.app.handlers.render_complete, bpy.app.handlers.render_cancel):
        while handle_render_done_cb in ls:
            ls.remove(handle_render_done_cb)


def handle_render_done_cb(self):
    global_state["is_render"] = True


def handle_render_init():
    handle_render_clear()
    bpy.app.handlers.render_complete.append(handle_render_done_cb)
    bpy.app.handlers.render_cancel.append(handle_render_done_cb)
    global_state["is_render"] = False


def demo_mode_auto_select():

    play_area = 0
    render_area = 0

    totimg = 0

    for area in bpy.context.window.screen.areas:
        size = area.width * area.height
        if area.type in {'VIEW_3D', 'GRAPH_EDITOR', 'DOPESHEET_EDITOR', 'NLA_EDITOR', 'TIMELINE'}:
            play_area += size
        elif area.type in {'IMAGE_EDITOR', 'SEQUENCE_EDITOR', 'NODE_EDITOR'}:
            render_area += size

        if area.type == 'IMAGE_EDITOR':
            totimg += 1

    # since our test files have this as defacto standard
    scene = bpy.context.scene
    if totimg >= 2 and (scene.camera or scene.render.use_sequencer):
        mode = 'RENDER'
    else:
        if play_area >= render_area:
            mode = 'PLAY'
        else:
            mode = 'RENDER'

    if 0:
        return 'PLAY'

    return mode


def demo_mode_next_file(step=1):

    # support for temp
    if global_config_files[global_state["demo_index"]].get("is_tmp"):
        del global_config_files[global_state["demo_index"]]
        global_state["demo_index"] -= 1

    demo_index_next = (global_state["demo_index"] + step) % len(global_config_files)

    if global_state["exit"] and step > 0:
        # check if we cycled
        if demo_index_next < global_state["demo_index"]:
            import sys
            sys.exit(0)

    global_state["demo_index"] = demo_index_next
    print(global_state["demo_index"], "....")
    print("func:demo_mode_next_file", global_state["demo_index"], "of", len(global_config_files))
    filepath = global_config_files[global_state["demo_index"]]["file"]
    bpy.ops.wm.open_mainfile(filepath=filepath)


def demo_mode_timer_add():
    global_state["timer"] = bpy.context.window_manager.event_timer_add(0.8, window=bpy.context.window)


def demo_mode_timer_remove():
    if global_state["timer"]:
        bpy.context.window_manager.event_timer_remove(global_state["timer"])
        global_state["timer"] = None


def demo_mode_load_file():
    """ Take care, this can only do limited functions since its running
        before the file is fully loaded.
        Some operators will crash like playing an animation.
    """
    print("func:demo_mode_load_file")
    DemoMode.first_run = True
    bpy.ops.wm.demo_mode('EXEC_DEFAULT')


def demo_mode_temp_file():
    """ Initialize a temp config for the duration of the play time.
        Use this so we can initialize the demo intro screen but not show again.
    """
    assert (global_state["demo_index"] == 0)

    temp_config = global_config_fallback.copy()
    temp_config["anim_time_min"] = 0.0
    temp_config["anim_time_max"] = 60.0
    temp_config["anim_cycles"] = 0  # ensures we switch when hitting the end
    temp_config["mode"] = 'PLAY'
    temp_config["is_tmp"] = True

    global_config_files.insert(0, temp_config)


def demo_mode_init():
    print("func:demo_mode_init")
    DemoKeepAlive.ensure()

    if 1:
        global_config.clear()
        global_config.update(global_config_files[global_state["demo_index"]])

    print(global_config)

    demo_mode_timer_add()

    if global_config["mode"] == 'AUTO':
        global_config["mode"] = demo_mode_auto_select()

    if global_config["mode"] == 'PLAY':
        global_state["last_frame"] = -1
        global_state["anim_cycles"] = 0
        bpy.ops.screen.animation_play()

    elif global_config["mode"] == 'RENDER':
        print("  render")

        # setup scene.
        scene = bpy.context.scene
        scene.render.filepath = "TEMP_RENDER"
        scene.render.image_settings.file_format = 'AVI_JPEG' if global_config["anim_render"] else 'PNG'
        scene.render.use_file_extension = False
        scene.render.use_placeholder = False
        try:
            # XXX - without this rendering will crash because of a bug in blender!
            bpy.ops.wm.redraw_timer(type='DRAW_WIN_SWAP', iterations=1)
            if global_config["anim_render"]:
                bpy.ops.render.render('INVOKE_DEFAULT', animation=True)
            else:
                bpy.ops.render.render('INVOKE_DEFAULT')  # write_still=True, no need to write now.

                handle_render_init()

        except RuntimeError:  # no camera for eg:
            import traceback
            traceback.print_exc()

    else:
        raise Exception("Unsupported mode %r" % global_config["mode"])

    global_state["init_time"] = global_state["last_switch"] = time.time()
    global_state["render_time"] = -1.0


def demo_mode_update():
    time_current = time.time()
    time_delta = time_current - global_state["last_switch"]
    time_total = time_current - global_state["init_time"]

    # --------------------------------------------------------------------------
    # ANIMATE MODE
    if global_config["mode"] == 'PLAY':
        frame = bpy.context.scene.frame_current
        # check for exit
        if time_total > global_config["anim_time_max"]:
            demo_mode_next_file()
            return
        # above cycles and minimum display time
        if (
                (time_total > global_config["anim_time_min"]) and
                (global_state["anim_cycles"] > global_config["anim_cycles"])
        ):
            # looped enough now.
            demo_mode_next_file()
            return

        # run update funcs
        if global_state["reset_anim"]:
            global_state["reset_anim"] = False
            bpy.ops.screen.animation_cancel(restore_frame=False)
            bpy.ops.screen.animation_play()

        # warning, switching the screen can switch the scene
        # and mess with our last-frame/cycles counting.
        if global_config["anim_screen_switch"]:
            # print(time_delta, 1)
            if time_delta > global_config["anim_screen_switch"]:
                window = bpy.context.window
                scene = window.scene
                workspace = window.workspace
                index = bpy.data.workspaces.keys().index(workspace.name)
                workspace_new = bpy.data.workspaces[(index + 1) % len(bpy.data.workspaces)]
                window.workspace = workspace_new

                global_state["last_switch"] = time_current

                # If we also switch scenes then reset last frame
                # otherwise it could mess up cycle calculation.
                if scene != window.scene:
                    global_state["last_frame"] = -1

                # if global_config["mode"] == 'PLAY':
                if 1:
                    global_state["reset_anim"] = True

        # did we loop?
        if global_state["last_frame"] > frame:
            print("Cycle!")
            global_state["anim_cycles"] += 1

        global_state["last_frame"] = frame

    # --------------------------------------------------------------------------
    # RENDER MODE
    elif global_config["mode"] == 'RENDER':
        if global_state["is_render"]:
            # wait until the time has passed
            # XXX, todo, if rendering an anim we need some way to check its done.
            if global_state["render_time"] == -1.0:
                global_state["render_time"] = time.time()
            else:
                if time.time() - global_state["render_time"] > global_config["display_render"]:
                    handle_render_clear()
                    demo_mode_next_file()
                    return
    else:
        raise Exception("Unsupported mode %r" % global_config["mode"])

# -----------------------------------------------------------------------------
# modal operator


class DemoKeepAlive:
    secret_attr = "_keepalive"

    @staticmethod
    def ensure():
        if DemoKeepAlive.secret_attr not in bpy.app.driver_namespace:
            bpy.app.driver_namespace[DemoKeepAlive.secret_attr] = DemoKeepAlive()

    @staticmethod
    def remove():
        if DemoKeepAlive.secret_attr in bpy.app.driver_namespace:
            del bpy.app.driver_namespace[DemoKeepAlive.secret_attr]

    def __del__(self):
        """ Hack, when the file is loaded the drivers namespace is cleared.
        """
        if DemoMode.enabled:
            demo_mode_load_file()


class DemoMode(bpy.types.Operator):
    bl_idname = "wm.demo_mode"
    bl_label = "Demo"

    enabled = False
    first_run = True

    def cleanup(self, disable=False):
        demo_mode_timer_remove()
        DemoMode.first_run = True

        if disable:
            DemoMode.enabled = False
            DemoKeepAlive.remove()

    def modal(self, context, event):
        # print("DemoMode.modal", global_state["anim_cycles"])
        if not DemoMode.enabled:
            self.cleanup(disable=True)
            return {'CANCELLED'}

        if event.type == 'ESC':
            self.cleanup(disable=True)
            # disable here and not in cleanup because this is a user level disable.
            # which should stay disabled until explicitly enabled again.
            return {'CANCELLED'}

        # print(event.type)
        if DemoMode.first_run:
            DemoMode.first_run = False

            demo_mode_init()
        else:
            demo_mode_update()

        return {'PASS_THROUGH'}

    def execute(self, context):
        print("func:DemoMode.execute:", len(global_config_files), "files")

        use_temp = False

        # load config if not loaded
        if not global_config_files:
            load_config()
            use_temp = True

        if not global_config_files:
            self.report(
                {'INFO'},
                "No configuration found with text or file: %s. Run File -> Demo Mode Setup" %
                DEMO_CFG,
            )
            return {'CANCELLED'}

        if use_temp:
            demo_mode_temp_file()  # play this once through then never again

        # toggle
        if DemoMode.enabled and DemoMode.first_run is False:
            # this actually cancells the previous running instance
            # should never happen now, DemoModeControl is for this.
            return {'CANCELLED'}
        else:
            DemoMode.enabled = True

            context.window_manager.modal_handler_add(self)
            return {'RUNNING_MODAL'}

    def cancel(self, context):
        print("func:DemoMode.cancel")
        # disable here means no running on file-load.
        self.cleanup()

    # call from DemoModeControl
    @classmethod
    def disable(cls):
        if cls.enabled and cls.first_run is False:
            # this actually cancells the previous running instance
            # should never happen now, DemoModeControl is for this.
            cls.enabled = False


class DemoModeControl(bpy.types.Operator):
    bl_idname = "wm.demo_mode_control"
    bl_label = "Control"

    mode: bpy.props.EnumProperty(
        items=(
            ('PREV', "Prev", ""),
            ('PAUSE', "Pause", ""),
            ('NEXT', "Next", ""),
        ),
        name="Mode"
    )

    def execute(self, context):
        mode = self.mode
        if mode == 'PREV':
            demo_mode_next_file(-1)
        elif mode == 'NEXT':
            demo_mode_next_file(1)
        else:  # pause
            DemoMode.disable()
        return {'FINISHED'}


def menu_func(self, context):
    # print("func:menu_func - DemoMode.enabled:", DemoMode.enabled, "bpy.app.driver_namespace:", DemoKeepAlive.secret_attr not in bpy.app.driver_namespace, 'global_state["timer"]:', global_state["timer"])
    layout = self.layout
    layout.operator_context = 'EXEC_DEFAULT'
    row = layout.row(align=True)
    row.label(text="Demo Mode:")
    if not DemoMode.enabled:
        row.operator("wm.demo_mode", icon='PLAY', text="")
    else:
        row.operator("wm.demo_mode_control", icon='REW', text="").mode = 'PREV'
        row.operator("wm.demo_mode_control", icon='PAUSE', text="").mode = 'PAUSE'
        row.operator("wm.demo_mode_control", icon='FF', text="").mode = 'NEXT'


def register():
    bpy.utils.register_class(DemoMode)
    bpy.utils.register_class(DemoModeControl)
    bpy.types.INFO_HT_header.append(menu_func)


def unregister():
    bpy.utils.unregister_class(DemoMode)
    bpy.utils.unregister_class(DemoModeControl)
    bpy.types.INFO_HT_header.remove(menu_func)


# -----------------------------------------------------------------------------
# parse args

def load_config(cfg_name=DEMO_CFG):
    namespace = {}
    global_config_files.clear()
    basedir = os.path.dirname(bpy.data.filepath)

    text = bpy.data.texts.get(cfg_name)
    if text is None:
        demo_path = os.path.join(basedir, cfg_name)
        if os.path.exists(demo_path):
            print("Using config file: %r" % demo_path)
            demo_file = open(demo_path, "r")
            demo_data = demo_file.read()
            demo_file.close()
        else:
            demo_data = ""
    else:
        print("Using config textblock: %r" % cfg_name)
        demo_data = text.as_string()
        demo_path = os.path.join(bpy.data.filepath, cfg_name)  # fake

    if not demo_data:
        print("Could not find %r textblock or %r file." % (DEMO_CFG, demo_path))
        return False

    namespace["__file__"] = demo_path

    exec(demo_data, namespace, namespace)

    demo_config = namespace["config"]
    demo_search_path = namespace.get("search_path")
    global_state["exit"] = namespace.get("exit", False)

    if demo_search_path is None:
        print("reading: %r, no search_path found, missing files wont be searched." % demo_path)
    if demo_search_path.startswith("//"):
        demo_search_path = bpy.path.abspath(demo_search_path)
    if not os.path.exists(demo_search_path):
        print("reading: %r, search_path %r does not exist." % (demo_path, demo_search_path))
        demo_search_path = None

    blend_lookup = {}
    # initialize once, case insensitive dict

    def lookup_file(filepath):
        filename = os.path.basename(filepath).lower()

        if not blend_lookup:
            # ensure only ever run once.
            blend_lookup[None] = None

            def blend_dict_items(path):
                for dirpath, dirnames, filenames in os.walk(path):
                    # skip '.git'
                    dirnames[:] = [d for d in dirnames if not d.startswith(".")]
                    for filename in filenames:
                        if filename.lower().endswith(".blend"):
                            filepath = os.path.join(dirpath, filename)
                            yield (filename.lower(), filepath)

            blend_lookup.update(dict(blend_dict_items(demo_search_path)))

        # fallback to original file
        return blend_lookup.get(filename, filepath)
    # done with search lookup

    for filecfg in demo_config:
        filepath_test = filecfg["file"]
        if not os.path.exists(filepath_test):
            filepath_test = os.path.join(basedir, filecfg["file"])
        if not os.path.exists(filepath_test):
            filepath_test = lookup_file(filepath_test)  # attempt to get from searchpath
        if not os.path.exists(filepath_test):
            print("Can't find %r or %r, skipping!")
            continue

        filecfg["file"] = os.path.normpath(filepath_test)

        # sanitize
        filecfg["file"] = os.path.abspath(filecfg["file"])
        filecfg["file"] = os.path.normpath(filecfg["file"])
        print("  Adding: %r" % filecfg["file"])
        global_config_files.append(filecfg)

    print("found %d files" % len(global_config_files))

    global_state["basedir"] = basedir

    return bool(global_config_files)


# support direct execution
if __name__ == "__main__":
    register()

    demo_mode_load_file()  # kick starts the modal operator



# SPDX-FileCopyrightText: 2016-2020 by Nathan Lovato, Daniel Oakey, Razvan Radulescu, and contributors
#
# SPDX-License-Identifier: GPL-3.0-or-later

import bpy
import operator

from .utils.global_settings import SequenceTypes
from .utils.functions import convert_duration_to_frames
from .utils.doc import doc_name, doc_idname, doc_brief, doc_description


class POWER_SEQUENCER_OT_make_hold_frame(bpy.types.Operator):
    """
    *brief* Make a hold frame from the active strip

    Converts the image under the cursor to a hold frame, to create a pause effect in the video,
    using the active sequence
    """

    doc = {
        "name": doc_name(__qualname__),
        "demo": "",
        "description": doc_description(__doc__),
        "shortcuts": [],
        "keymap": "Sequencer",
    }
    bl_idname = doc_idname(__qualname__)
    bl_label = doc["name"]
    bl_description = doc_brief(doc["description"])
    bl_options = {"REGISTER", "UNDO"}

    strip_duration: bpy.props.FloatProperty(
        name="Strip Duration",
        description="The duration in seconds of the new strip, if 0.0 it will use the gap as its duration",
        default=0.0,
        min=0.0,
    )

    @classmethod
    def poll(cls, context):
        return context.scene.sequence_editor.active_strip.type in SequenceTypes.VIDEO

    def invoke(self, context, event):
        window_manager = context.window_manager
        return window_manager.invoke_props_dialog(self)

    def execute(self, context):
        scene = context.scene
        active = scene.sequence_editor.active_strip
        sequencer = bpy.ops.sequencer
        transform = bpy.ops.transform

        frame_start = scene.frame_current

        if not active.frame_final_start <= frame_start < active.frame_final_end:
            return {"FINISHED"}

        if frame_start == active.frame_final_start:
            scene.frame_current = frame_start + 1

        # Detect the gap automatically
        offset = convert_duration_to_frames(context, self.strip_duration)
        if self.strip_duration <= 0.0:
            try:
                next_strip_start = next(
                    s
                    for s in sorted(context.sequences, key=operator.attrgetter("frame_final_start"))
                    if s.frame_final_start > active.frame_final_end
                ).frame_final_start
                offset = next_strip_start - active.frame_final_end
            except Exception:
                pass

        active.select = True
        source_blend_type = active.blend_type
        sequencer.split(frame=scene.frame_current, type="SOFT", side="RIGHT")
        transform.seq_slide(value=(offset, 0))
        sequencer.split(frame=scene.frame_current + offset + 1, type="SOFT", side="LEFT")
        transform.seq_slide(value=(-offset, 0))

        sequencer.meta_make()
        active = scene.sequence_editor.active_strip
        active.name = "Hold frame"
        active.blend_type = source_blend_type
        active.select_right_handle = True
        transform.seq_slide(value=(offset, 0))

        scene.frame_current = frame_start

        active.select = True
        active.select_right_handle = False
        active.select_left_handle = False
        return {"FINISHED"}



# SPDX-FileCopyrightText: 2019-2023 Blender Foundation
#
# SPDX-License-Identifier: GPL-2.0-or-later

import gpu
from mathutils import Vector, Matrix


class SnapDrawn():
    __slots__ = (
        'out_color',
        'face_color',
        'edge_color',
        'vert_color',
        'center_color',
        'perpendicular_color',
        'constrain_shift_color',
        'axis_x_color',
        'axis_y_color',
        'axis_z_color',
        'rv3d',
        '_point_size',
        '_line_width',
        '_format_pos',
        '_format_pos_and_color',
        '_program_unif_col',
        '_program_smooth_col',
        '_UBO',
        '_batch_point',)

    def __init__(self, out_color, face_color,
                 edge_color, vert_color, center_color,
                 perpendicular_color, constrain_shift_color,
                 axis_x_color, axis_y_color, axis_z_color, rv3d, ui_scale):

        import gpu

        self.out_color = out_color
        self.face_color = face_color
        self.edge_color = edge_color
        self.vert_color = vert_color
        self.center_color = center_color
        self.perpendicular_color = perpendicular_color
        self.constrain_shift_color = constrain_shift_color

        self.axis_x_color = axis_x_color
        self.axis_y_color = axis_y_color
        self.axis_z_color = axis_z_color

        self.rv3d = rv3d

        self._point_size = 5 * ui_scale
        self._line_width = 3 * ui_scale

        self._format_pos = gpu.types.GPUVertFormat()
        self._format_pos.attr_add(
            id="pos", comp_type='F32', len=3, fetch_mode='FLOAT')

        self._format_pos_and_color = gpu.types.GPUVertFormat()
        self._format_pos_and_color.attr_add(
            id="pos", comp_type='F32', len=3, fetch_mode='FLOAT')
        self._format_pos_and_color.attr_add(
            id="color", comp_type='F32', len=4, fetch_mode='FLOAT')

        self._UBO = None

        self._batch_point = None

    def _gl_state_push(self, ob_mat=None):
        clip_planes = self.rv3d.clip_planes if self.rv3d.use_clip_planes else None
        config = 'CLIPPED' if clip_planes else 'DEFAULT'
        self._program_unif_col = gpu.shader.from_builtin(
            "UNIFORM_COLOR", config=config)
        self._program_smooth_col = gpu.shader.from_builtin(
            "SMOOTH_COLOR", config=config)

        gpu.state.program_point_size_set(False)
        gpu.state.blend_set('ALPHA')
        gpu.matrix.push()
        if ob_mat:
            gpu.matrix.multiply_matrix(ob_mat)

        if clip_planes:
            gpu.state.clip_distances_set(4)
            if self._UBO is None:
                import ctypes

                class _GPUClipPlanes(ctypes.Structure):
                    _pack_ = 16
                    _fields_ = [
                        ("ModelMatrix", (ctypes.c_float * 4) * 4),
                        ("world", (ctypes.c_float * 4) * 6),
                    ]

                mat = ob_mat.transposed() if ob_mat else Matrix.Identity(4)

                UBO_data = _GPUClipPlanes()
                UBO_data.ModelMatrix[0] = mat[0][:]
                UBO_data.ModelMatrix[1] = mat[1][:]
                UBO_data.ModelMatrix[2] = mat[2][:]
                UBO_data.ModelMatrix[3] = mat[3][:]

                UBO_data.world[0] = clip_planes[0][:]
                UBO_data.world[1] = clip_planes[1][:]
                UBO_data.world[2] = clip_planes[2][:]
                UBO_data.world[3] = clip_planes[3][:]

                self._UBO = gpu.types.GPUUniformBuf(UBO_data)

            self._program_unif_col.bind()
            self._program_unif_col.uniform_block("clipPlanes", self._UBO)

            self._program_smooth_col.bind()
            self._program_smooth_col.uniform_block("clipPlanes", self._UBO)

    def _gl_state_restore(self):
        gpu.state.blend_set('NONE')
        gpu.matrix.pop()
        if self._UBO:
            # del self._UBO
            self._UBO = None
            gpu.state.clip_distances_set(0)

    def batch_line_strip_create(self, coords):
        from gpu.types import (
            GPUVertBuf,
            GPUBatch,
        )

        vbo = GPUVertBuf(self._format_pos, len=len(coords))
        vbo.attr_fill(0, data=coords)
        batch_lines = GPUBatch(type="LINE_STRIP", buf=vbo)
        return batch_lines

    def batch_lines_smooth_color_create(self, coords, colors):
        from gpu.types import (
            GPUVertBuf,
            GPUBatch,
        )

        vbo = GPUVertBuf(self._format_pos_and_color, len=len(coords))
        vbo.attr_fill(0, data=coords)
        vbo.attr_fill(1, data=colors)
        batch_lines = GPUBatch(type="LINES", buf=vbo)
        return batch_lines

    def batch_triangles_create(self, coords):
        from gpu.types import (
            GPUVertBuf,
            GPUBatch,
        )

        vbo = GPUVertBuf(self._format_pos, len=len(coords))
        vbo.attr_fill(0, data=coords)
        batch_tris = GPUBatch(type="TRIS", buf=vbo)
        return batch_tris

    def batch_point_get(self):
        if self._batch_point is None:
            from gpu.types import (
                GPUVertBuf,
                GPUBatch,
            )
            vbo = GPUVertBuf(self._format_pos, len=1)
            vbo.attr_fill(0, ((0.0, 0.0, 0.0),))
            self._batch_point = GPUBatch(type="POINTS", buf=vbo)
        return self._batch_point

    def draw(self, type, location, list_verts_co, vector_constrain, prevloc):
        import gpu

        self._gl_state_push()
        self._program_unif_col.bind()

        if list_verts_co:
            # draw 3d line OpenGL in the 3D View
            winmat = gpu.matrix.get_projection_matrix()
            winmat[3][2] -= 0.0001
            gpu.matrix.push_projection()
            gpu.matrix.load_projection_matrix(winmat)
            gpu.state.line_width_set(self._line_width)

            batch = self.batch_line_strip_create(
                [v.to_tuple() for v in list_verts_co] + [location.to_tuple()])

            self._program_unif_col.bind()
            self._program_unif_col.uniform_float("color", (1.0, 0.8, 0.0, 0.5))

            batch.draw(self._program_unif_col)
            gpu.matrix.pop_projection()
            del batch

        gpu.state.depth_test_set('NONE')

        point_batch = self.batch_point_get()
        if vector_constrain:
            if prevloc:
                gpu.state.point_size_set(self._point_size)
                gpu.matrix.translate(prevloc)

                self._program_unif_col.bind()
                self._program_unif_col.uniform_float(
                    "color", (1.0, 1.0, 1.0, 0.5))

                point_batch.draw(self._program_unif_col)
                gpu.matrix.translate(-prevloc)

            if vector_constrain[2] == 'X':
                Color4f = self.axis_x_color
            elif vector_constrain[2] == 'Y':
                Color4f = self.axis_y_color
            elif vector_constrain[2] == 'Z':
                Color4f = self.axis_z_color
            else:
                Color4f = self.constrain_shift_color
        else:
            if type == 'OUT':
                Color4f = self.out_color
            elif type == 'FACE':
                Color4f = self.face_color
            elif type == 'EDGE':
                Color4f = self.edge_color
            elif type == 'VERT':
                Color4f = self.vert_color
            elif type == 'CENTER':
                Color4f = self.center_color
            elif type == 'PERPENDICULAR':
                Color4f = self.perpendicular_color
            else:  # type == None
                Color4f = self.out_color

        gpu.state.point_size_set(2 * self._point_size)

        gpu.matrix.translate(location)
        self._program_unif_col.bind()
        self._program_unif_col.uniform_float("color", Color4f)
        point_batch.draw(self._program_unif_col)

        # restore opengl defaults
        gpu.state.point_size_set(1.0)
        gpu.state.line_width_set(1.0)
        gpu.state.depth_test_set('LESS_EQUAL')

        self._gl_state_restore()

    def draw_elem(self, snap_obj, bm, elem):
        # TODO: Cache coords (because antialiasing)
        import gpu
        from bmesh.types import (
            BMVert,
            BMEdge,
            BMFace,
        )

        self._gl_state_push(snap_obj.mat)
        gpu.state.depth_test_set('NONE')

        if isinstance(elem, BMVert):
            if elem.link_edges:
                import numpy as np

                color = self.vert_color
                edges = np.empty((len(elem.link_edges), 2), [
                                 ("pos", "f4", 3), ("color", "f4", 4)])
                edges["pos"][:, 0] = elem.co
                edges["pos"][:, 1] = [e.other_vert(
                    elem).co for e in elem.link_edges]
                edges["color"][:, 0] = color
                edges["color"][:, 1] = (color[0], color[1], color[2], 0.0)
                edges.shape = -1

                self._program_smooth_col.bind()
                gpu.state.line_width_set(self._line_width)
                batch = self.batch_lines_smooth_color_create(
                    edges["pos"], edges["color"])
                batch.draw(self._program_smooth_col)
                gpu.state.line_width_set(1.0)
        else:
            if isinstance(elem, BMEdge):
                self._program_unif_col.bind()
                self._program_unif_col.uniform_float("color", self.edge_color)

                gpu.state.line_width_set(self._line_width)
                batch = self.batch_line_strip_create(
                    [v.co for v in elem.verts])
                batch.draw(self._program_unif_col)
                gpu.state.line_width_set(1.0)

            elif isinstance(elem, BMFace):
                if len(snap_obj.data) == 2:
                    face_color = self.face_color[0], self.face_color[1], self.face_color[2], self.face_color[3] * 0.2
                    self._program_unif_col.bind()
                    self._program_unif_col.uniform_float("color", face_color)

                    tris = snap_obj.data[1].get_loop_tri_co_by_bmface(bm, elem)
                    tris.shape = (-1, 3)
                    batch = self.batch_triangles_create(tris)
                    batch.draw(self._program_unif_col)

        # restore opengl defaults
        gpu.state.depth_test_set('LESS_EQUAL')

        self._gl_state_restore()



# SPDX-FileCopyrightText: 2019-2022 Blender Foundation
#
# SPDX-License-Identifier: GPL-2.0-or-later

import bpy
from . import mathematics
from . import curves
from . import util

from mathutils import Vector

algoPOV = None
algoDIR = None


class BezierSegmentIntersectionPoint:
    def __init__(self, segment, parameter, intersectionPoint):
        self.segment = segment
        self.parameter = parameter
        self.intersectionPoint = intersectionPoint


class BezierSegmentsIntersector:
    def __init__(self, segment1, segment2, worldMatrix1, worldMatrix2):
        self.segment1 = segment1
        self.segment2 = segment2
        self.worldMatrix1 = worldMatrix1
        self.worldMatrix2 = worldMatrix2

    def CalcFirstIntersection(self, nrSamples1, nrSamples2):
        algorithm = bpy.context.scene.curvetools.IntersectCurvesAlgorithm

        if algorithm == '3D':
            return self.CalcFirstRealIntersection3D(nrSamples1, nrSamples2)

        if algorithm == 'From_View':
            global algoDIR
            if algoDIR is not None:
                return self.CalcFirstRealIntersectionFromViewDIR(nrSamples1, nrSamples2)

            global algoPOV
            if algoPOV is not None:
                return self.CalcFirstRealIntersectionFromViewPOV(nrSamples1, nrSamples2)

        return None

    def CalcFirstIntersection3D(self, nrSamples1, nrSamples2):
        fltNrSamples1 = float(nrSamples1)
        fltNrSamples2 = float(nrSamples2)

        limitDistance = bpy.context.scene.curvetools.LimitDistance

        for iSample1 in range(nrSamples1):
            segPar10 = float(iSample1) / fltNrSamples1
            segPar11 = float(iSample1 + 1) / fltNrSamples1
            P0 = self.worldMatrix1 @ self.segment1.CalcPoint(parameter=segPar10)
            P1 = self.worldMatrix1 @ self.segment1.CalcPoint(parameter=segPar11)

            for iSample2 in range(nrSamples2):
                segPar20 = float(iSample2) / fltNrSamples2
                segPar21 = float(iSample2 + 1) / fltNrSamples2
                Q0 = self.worldMatrix2 @ self.segment2.CalcPoint(parameter=segPar20)
                Q1 = self.worldMatrix2 @ self.segment2.CalcPoint(parameter=segPar21)

                intersectionPointData = mathematics.CalcIntersectionPointLineSegments(P0, P1, Q0, Q1, limitDistance)
                if intersectionPointData is None:
                    continue

                intersectionSegment1Parameter = segPar10 + (intersectionPointData[0] / fltNrSamples1)
                intersectionPoint1 = BezierSegmentIntersectionPoint(self.segment1,
                                                                    intersectionSegment1Parameter,
                                                                    intersectionPointData[2])

                intersectionSegment2Parameter = segPar20 + (intersectionPointData[1] / fltNrSamples2)
                intersectionPoint2 = BezierSegmentIntersectionPoint(self.segment2,
                                                                    intersectionSegment2Parameter,
                                                                    intersectionPointData[3])

                return [intersectionPoint1, intersectionPoint2]

        return None

    def CalcFirstRealIntersection3D(self, nrSamples1, nrSamples2):
        fltNrSamples1 = float(nrSamples1)
        fltNrSamples2 = float(nrSamples2)

        limitDistance = bpy.context.scene.curvetools.LimitDistance

        for iSample1 in range(nrSamples1):
            segPar10 = float(iSample1) / fltNrSamples1
            segPar11 = float(iSample1 + 1) / fltNrSamples1
            P0 = self.worldMatrix1 @ self.segment1.CalcPoint(parameter=segPar10)
            P1 = self.worldMatrix1 @ self.segment1.CalcPoint(parameter=segPar11)

            for iSample2 in range(nrSamples2):
                segPar20 = float(iSample2) / fltNrSamples2
                segPar21 = float(iSample2 + 1) / fltNrSamples2
                Q0 = self.worldMatrix2 @ self.segment2.CalcPoint(parameter=segPar20)
                Q1 = self.worldMatrix2 @ self.segment2.CalcPoint(parameter=segPar21)

                intersectionPointData = mathematics.CalcIntersectionPointLineSegments(P0, P1, Q0, Q1, limitDistance)
                if intersectionPointData is None:
                    continue

                # intersection point can't be an existing point
                intersectionSegment1Parameter = segPar10 + (intersectionPointData[0] / (fltNrSamples1))
                worldPoint1 = self.worldMatrix1 @ self.segment1.CalcPoint(parameter=intersectionSegment1Parameter)
                if (mathematics.IsSamePoint(P0, worldPoint1, limitDistance)) or \
                   (mathematics.IsSamePoint(P1, worldPoint1, limitDistance)):

                    intersectionPoint1 = None
                else:
                    intersectionPoint1 = BezierSegmentIntersectionPoint(self.segment1,
                                                                        intersectionSegment1Parameter,
                                                                        worldPoint1)

                intersectionSegment2Parameter = segPar20 + (intersectionPointData[1] / (fltNrSamples2))
                worldPoint2 = self.worldMatrix2 @ self.segment2.CalcPoint(parameter=intersectionSegment2Parameter)
                if (mathematics.IsSamePoint(Q0, worldPoint2, limitDistance)) or \
                   (mathematics.IsSamePoint(Q1, worldPoint2, limitDistance)):

                    intersectionPoint2 = None
                else:
                    intersectionPoint2 = BezierSegmentIntersectionPoint(self.segment2,
                                                                        intersectionSegment2Parameter,
                                                                        worldPoint2)

                return [intersectionPoint1, intersectionPoint2]

        return None

    def CalcFirstIntersectionFromViewDIR(self, nrSamples1, nrSamples2):
        global algoDIR

        fltNrSamples1 = float(nrSamples1)
        fltNrSamples2 = float(nrSamples2)

        for iSample1 in range(nrSamples1):
            segPar10 = float(iSample1) / fltNrSamples1
            segPar11 = float(iSample1 + 1) / fltNrSamples1
            P0 = self.worldMatrix1 @ self.segment1.CalcPoint(parameter=segPar10)
            P1 = self.worldMatrix1 @ self.segment1.CalcPoint(parameter=segPar11)

            for iSample2 in range(nrSamples2):
                segPar20 = float(iSample2) / fltNrSamples2
                segPar21 = float(iSample2 + 1) / fltNrSamples2
                Q0 = self.worldMatrix2 @ self.segment2.CalcPoint(parameter=segPar20)
                Q1 = self.worldMatrix2 @ self.segment2.CalcPoint(parameter=segPar21)

                intersectionPointData = mathematics.CalcIntersectionPointsLineSegmentsDIR(P0, P1, Q0, Q1, algoDIR)
                if intersectionPointData is None:
                    continue

                intersectionSegment1Parameter = segPar10 + (intersectionPointData[0] / fltNrSamples1)
                worldPoint1 = self.worldMatrix1 @ self.segment1.CalcPoint(parameter=intersectionSegment1Parameter)
                intersectionPoint1 = BezierSegmentIntersectionPoint(self.segment1,
                                                                    intersectionSegment1Parameter,
                                                                    worldPoint1)

                intersectionSegment2Parameter = segPar20 + (intersectionPointData[1] / fltNrSamples2)
                worldPoint2 = self.worldMatrix2 @ self.segment2.CalcPoint(parameter=intersectionSegment2Parameter)
                intersectionPoint2 = BezierSegmentIntersectionPoint(self.segment2,
                                                                    intersectionSegment2Parameter,
                                                                    worldPoint2)

                return [intersectionPoint1, intersectionPoint2]

        return None

    def CalcFirstRealIntersectionFromViewDIR(self, nrSamples1, nrSamples2):
        global algoDIR

        fltNrSamples1 = float(nrSamples1)
        fltNrSamples2 = float(nrSamples2)

        limitDistance = bpy.context.scene.curvetools.LimitDistance

        for iSample1 in range(nrSamples1):
            segPar10 = float(iSample1) / fltNrSamples1
            segPar11 = float(iSample1 + 1) / fltNrSamples1
            P0 = self.worldMatrix1 @ self.segment1.CalcPoint(parameter=segPar10)
            P1 = self.worldMatrix1 @ self.segment1.CalcPoint(parameter=segPar11)

            for iSample2 in range(nrSamples2):
                segPar20 = float(iSample2) / fltNrSamples2
                segPar21 = float(iSample2 + 1) / fltNrSamples2
                Q0 = self.worldMatrix2 @ self.segment2.CalcPoint(parameter=segPar20)
                Q1 = self.worldMatrix2 @ self.segment2.CalcPoint(parameter=segPar21)

                intersectionPointData = mathematics.CalcIntersectionPointsLineSegmentsDIR(P0, P1, Q0, Q1, algoDIR)
                if intersectionPointData is None:
                    continue

                # intersection point can't be an existing point
                intersectionSegment1Parameter = segPar10 + (intersectionPointData[0] / (fltNrSamples1))
                worldPoint1 = self.worldMatrix1 @ self.segment1.CalcPoint(parameter=intersectionSegment1Parameter)
                if (mathematics.IsSamePoint(P0, worldPoint1, limitDistance)) or \
                   (mathematics.IsSamePoint(P1, worldPoint1, limitDistance)):

                    intersectionPoint1 = None
                else:
                    intersectionPoint1 = BezierSegmentIntersectionPoint(self.segment1,
                                                                        intersectionSegment1Parameter,
                                                                        worldPoint1)

                intersectionSegment2Parameter = segPar20 + (intersectionPointData[1] / (fltNrSamples2))
                worldPoint2 = self.worldMatrix2 @ self.segment2.CalcPoint(parameter=intersectionSegment2Parameter)
                if (mathematics.IsSamePoint(Q0, worldPoint2, limitDistance)) or \
                   (mathematics.IsSamePoint(Q1, worldPoint2, limitDistance)):

                    intersectionPoint2 = None
                else:
                    intersectionPoint2 = BezierSegmentIntersectionPoint(self.segment2,
                                                                        intersectionSegment2Parameter,
                                                                        worldPoint2)

                return [intersectionPoint1, intersectionPoint2]

        return None

    def CalcFirstIntersectionFromViewPOV(self, nrSamples1, nrSamples2):
        global algoPOV

        fltNrSamples1 = float(nrSamples1)
        fltNrSamples2 = float(nrSamples2)

        for iSample1 in range(nrSamples1):
            segPar10 = float(iSample1) / fltNrSamples1
            segPar11 = float(iSample1 + 1) / fltNrSamples1
            P0 = self.worldMatrix1 @ self.segment1.CalcPoint(parameter=segPar10)
            P1 = self.worldMatrix1 @ self.segment1.CalcPoint(parameter=segPar11)

            for iSample2 in range(nrSamples2):
                segPar20 = float(iSample2) / fltNrSamples2
                segPar21 = float(iSample2 + 1) / fltNrSamples2
                Q0 = self.worldMatrix2 @ self.segment2.CalcPoint(parameter=segPar20)
                Q1 = self.worldMatrix2 @ self.segment2.CalcPoint(parameter=segPar21)

                intersectionPointData = mathematics.CalcIntersectionPointsLineSegmentsPOV(P0, P1, Q0, Q1, algoPOV)
                if intersectionPointData is None:
                    continue

                intersectionSegment1Parameter = segPar10 + (intersectionPointData[0] / fltNrSamples1)
                worldPoint1 = self.worldMatrix1 @ self.segment1.CalcPoint(parameter=intersectionSegment1Parameter)
                intersectionPoint1 = BezierSegmentIntersectionPoint(self.segment1,
                                                                    intersectionSegment1Parameter,
                                                                    worldPoint1)

                intersectionSegment2Parameter = segPar20 + (intersectionPointData[1] / fltNrSamples2)
                worldPoint2 = self.worldMatrix2 @ self.segment2.CalcPoint(parameter=intersectionSegment2Parameter)
                intersectionPoint2 = BezierSegmentIntersectionPoint(self.segment2,
                                                                    intersectionSegment2Parameter,
                                                                    worldPoint2)

                return [intersectionPoint1, intersectionPoint2]

        return None

    def CalcFirstRealIntersectionFromViewPOV(self, nrSamples1, nrSamples2):
        global algoPOV

        fltNrSamples1 = float(nrSamples1)
        fltNrSamples2 = float(nrSamples2)

        limitDistance = bpy.context.scene.curvetools.LimitDistance

        for iSample1 in range(nrSamples1):
            segPar10 = float(iSample1) / fltNrSamples1
            segPar11 = float(iSample1 + 1) / fltNrSamples1
            P0 = self.worldMatrix1 @ self.segment1.CalcPoint(parameter=segPar10)
            P1 = self.worldMatrix1 @ self.segment1.CalcPoint(parameter=segPar11)

            for iSample2 in range(nrSamples2):
                segPar20 = float(iSample2) / fltNrSamples2
                segPar21 = float(iSample2 + 1) / fltNrSamples2
                Q0 = self.worldMatrix2 @ self.segment2.CalcPoint(parameter=segPar20)
                Q1 = self.worldMatrix2 @ self.segment2.CalcPoint(parameter=segPar21)

                intersectionPointData = mathematics.CalcIntersectionPointsLineSegmentsPOV(P0, P1, Q0, Q1, algoPOV)
                if intersectionPointData is None:
                    continue

                # intersection point can't be an existing point
                intersectionSegment1Parameter = segPar10 + (intersectionPointData[0] / fltNrSamples1)
                worldPoint1 = self.worldMatrix1 @ self.segment1.CalcPoint(parameter=intersectionSegment1Parameter)
                if (mathematics.IsSamePoint(P0, worldPoint1, limitDistance)) or \
                   (mathematics.IsSamePoint(P1, worldPoint1, limitDistance)):

                    intersectionPoint1 = None
                else:
                    intersectionPoint1 = BezierSegmentIntersectionPoint(self.segment1,
                                                                        intersectionSegment1Parameter,
                                                                        worldPoint1)

                intersectionSegment2Parameter = segPar20 + (intersectionPointData[1] / fltNrSamples2)
                worldPoint2 = self.worldMatrix2 @ self.segment2.CalcPoint(parameter=intersectionSegment2Parameter)
                if (mathematics.IsSamePoint(Q0, worldPoint2, limitDistance)) or \
                   (mathematics.IsSamePoint(Q1, worldPoint2, limitDistance)):

                    intersectionPoint2 = None
                else:
                    intersectionPoint2 = BezierSegmentIntersectionPoint(self.segment2,
                                                                        intersectionSegment2Parameter,
                                                                        worldPoint2)

                return [intersectionPoint1, intersectionPoint2]

        return None

    def CalcIntersections(self, nrSamples1, nrSamples2):
        algorithm = bpy.context.scene.curvetools.IntersectCurvesAlgorithm

        if algorithm == '3D':
            return self.CalcIntersections3D(nrSamples1, nrSamples2)

        if algorithm == 'From_View':
            global algoDIR
            if algoDIR is not None:
                return self.CalcIntersectionsFromViewDIR(nrSamples1, nrSamples2)

            global algoPOV
            if algoPOV is not None:
                return self.CalcIntersectionsFromViewPOV(nrSamples1, nrSamples2)

        return [[], []]

    def CalcIntersections3D(self, nrSamples1, nrSamples2):
        rvIntersections1 = []
        rvIntersections2 = []

        fltNrSamples1 = float(nrSamples1)
        fltNrSamples2 = float(nrSamples2)

        limitDistance = bpy.context.scene.curvetools.LimitDistance

        for iSample1 in range(nrSamples1):
            segPar10 = float(iSample1) / fltNrSamples1
            segPar11 = float(iSample1 + 1) / fltNrSamples1
            P0 = self.worldMatrix1 @ self.segment1.CalcPoint(parameter=segPar10)
            P1 = self.worldMatrix1 @ self.segment1.CalcPoint(parameter=segPar11)

            for iSample2 in range(nrSamples2):
                segPar20 = float(iSample2) / fltNrSamples2
                segPar21 = float(iSample2 + 1) / fltNrSamples2
                Q0 = self.worldMatrix2 @ self.segment2.CalcPoint(parameter=segPar20)
                Q1 = self.worldMatrix2 @ self.segment2.CalcPoint(parameter=segPar21)

                intersectionPointData = mathematics.CalcIntersectionPointLineSegments(P0, P1, Q0, Q1, limitDistance)
                if intersectionPointData is None:
                    continue

                intersectionSegment1Parameter = segPar10 + (intersectionPointData[0] / fltNrSamples1)
                worldPoint1 = self.worldMatrix1 @ self.segment1.CalcPoint(parameter=intersectionSegment1Parameter)
                intersectionPoint1 = BezierSegmentIntersectionPoint(self.segment1,
                                                                    intersectionSegment1Parameter,
                                                                    worldPoint1)
                rvIntersections1.append(intersectionPoint1)

                intersectionSegment2Parameter = segPar20 + (intersectionPointData[1] / fltNrSamples2)
                worldPoint2 = self.worldMatrix2 @ self.segment2.CalcPoint(parameter=intersectionSegment2Parameter)
                intersectionPoint2 = BezierSegmentIntersectionPoint(self.segment2,
                                                                    intersectionSegment2Parameter,
                                                                    worldPoint2)
                rvIntersections2.append(intersectionPoint2)

        return [rvIntersections1, rvIntersections2]

    def CalcIntersectionsFromViewDIR(self, nrSamples1, nrSamples2):
        global algoDIR

        rvIntersections1 = []
        rvIntersections2 = []

        fltNrSamples1 = float(nrSamples1)
        fltNrSamples2 = float(nrSamples2)

        for iSample1 in range(nrSamples1):
            segPar10 = float(iSample1) / fltNrSamples1
            segPar11 = float(iSample1 + 1) / fltNrSamples1
            P0 = self.worldMatrix1 @ self.segment1.CalcPoint(parameter=segPar10)
            P1 = self.worldMatrix1 @ self.segment1.CalcPoint(parameter=segPar11)

            for iSample2 in range(nrSamples2):
                segPar20 = float(iSample2) / fltNrSamples2
                segPar21 = float(iSample2 + 1) / fltNrSamples2
                Q0 = self.worldMatrix2 @ self.segment2.CalcPoint(parameter=segPar20)
                Q1 = self.worldMatrix2 @ self.segment2.CalcPoint(parameter=segPar21)

                intersectionPointData = mathematics.CalcIntersectionPointsLineSegmentsDIR(P0, P1, Q0, Q1, algoDIR)
                if intersectionPointData is None:
                    continue

                intersectionSegment1Parameter = segPar10 + (intersectionPointData[0] / fltNrSamples1)
                worldPoint1 = self.worldMatrix1 @ self.segment1.CalcPoint(parameter=intersectionSegment1Parameter)
                intersectionPoint1 = BezierSegmentIntersectionPoint(self.segment1,
                                                                    intersectionSegment1Parameter,
                                                                    worldPoint1)
                rvIntersections1.append(intersectionPoint1)

                intersectionSegment2Parameter = segPar20 + (intersectionPointData[1] / fltNrSamples2)
                worldPoint2 = self.worldMatrix2 @ self.segment2.CalcPoint(parameter=intersectionSegment2Parameter)
                intersectionPoint2 = BezierSegmentIntersectionPoint(self.segment2,
                                                                    intersectionSegment2Parameter,
                                                                    worldPoint2)
                rvIntersections2.append(intersectionPoint2)

        return [rvIntersections1, rvIntersections2]

    def CalcIntersectionsFromViewPOV(self, nrSamples1, nrSamples2):
        global algoPOV

        rvIntersections1 = []
        rvIntersections2 = []

        fltNrSamples1 = float(nrSamples1)
        fltNrSamples2 = float(nrSamples2)

        for iSample1 in range(nrSamples1):
            segPar10 = float(iSample1) / fltNrSamples1
            segPar11 = float(iSample1 + 1) / fltNrSamples1
            P0 = self.worldMatrix1 @ self.segment1.CalcPoint(parameter=segPar10)
            P1 = self.worldMatrix1 @ self.segment1.CalcPoint(parameter=segPar11)

            for iSample2 in range(nrSamples2):
                segPar20 = float(iSample2) / fltNrSamples2
                segPar21 = float(iSample2 + 1) / fltNrSamples2
                Q0 = self.worldMatrix2 @ self.segment2.CalcPoint(parameter=segPar20)
                Q1 = self.worldMatrix2 @ self.segment2.CalcPoint(parameter=segPar21)

                intersectionPointData = mathematics.CalcIntersectionPointsLineSegmentsPOV(P0, P1, Q0, Q1, algoPOV)
                if intersectionPointData is None:
                    continue

                intersectionSegment1Parameter = segPar10 + (intersectionPointData[0] / fltNrSamples1)
                worldPoint1 = self.worldMatrix1 @ self.segment1.CalcPoint(parameter=intersectionSegment1Parameter)
                intersectionPoint1 = BezierSegmentIntersectionPoint(self.segment1,
                                                                    intersectionSegment1Parameter,
                                                                    worldPoint1)
                rvIntersections1.append(intersectionPoint1)

                intersectionSegment2Parameter = segPar20 + (intersectionPointData[1] / fltNrSamples2)
                worldPoint2 = self.worldMatrix2 @ self.segment2.CalcPoint(parameter=intersectionSegment2Parameter)
                intersectionPoint2 = BezierSegmentIntersectionPoint(self.segment2,
                                                                    intersectionSegment2Parameter,
                                                                    worldPoint2)
                rvIntersections2.append(intersectionPoint2)

        return [rvIntersections1, rvIntersections2]


class BezierSplineIntersectionPoint:
    def __init__(self, spline, bezierSegmentIntersectionPoint):
        self.spline = spline
        self.bezierSegmentIntersectionPoint = bezierSegmentIntersectionPoint


class BezierSplinesIntersector:
    def __init__(self, spline1, spline2, worldMatrix1, worldMatrix2):
        self.spline1 = spline1
        self.spline2 = spline2
        self.worldMatrix1 = worldMatrix1
        self.worldMatrix2 = worldMatrix2

    def CalcIntersections(self):
        rvIntersections1 = []
        rvIntersections2 = []

        try:
            nrSamplesPerSegment1 = int(self.spline1.resolution / self.spline1.nrSegments)
        except:
            nrSamplesPerSegment1 = 2
        if nrSamplesPerSegment1 < 2:
            nrSamplesPerSegment1 = 2

        try:
            nrSamplesPerSegment2 = int(self.spline2.resolution / self.spline2.nrSegments)
        except:
            nrSamplesPerSegment2 = 2
        if nrSamplesPerSegment2 < 2:
            nrSamplesPerSegment2 = 2

        for segment1 in self.spline1.segments:
            for segment2 in self.spline2.segments:
                segmentsIntersector = BezierSegmentsIntersector(segment1, segment2,
                                                                self.worldMatrix1, self.worldMatrix2)
                segmentIntersections = segmentsIntersector.CalcIntersections(nrSamplesPerSegment1, nrSamplesPerSegment2)
                if segmentIntersections is None:
                    continue

                segment1Intersections = segmentIntersections[0]
                for segmentIntersection in segment1Intersections:
                    splineIntersection = BezierSplineIntersectionPoint(self.spline1, segmentIntersection)
                    rvIntersections1.append(splineIntersection)

                segment2Intersections = segmentIntersections[1]
                for segmentIntersection in segment2Intersections:
                    splineIntersection = BezierSplineIntersectionPoint(self.spline2, segmentIntersection)
                    rvIntersections2.append(splineIntersection)

        return [rvIntersections1, rvIntersections2]


class CurvesIntersector:
    @staticmethod
    def FromSelection():
        selObjects = bpy.context.selected_objects
        if len(selObjects) != 2:
            raise Exception("len(selObjects) != 2")  # shouldn't be possible

        blenderActiveCurve = bpy.context.active_object
        blenderOtherCurve = selObjects[0]
        if blenderActiveCurve == blenderOtherCurve:
            blenderOtherCurve = selObjects[1]

        aCurve = curves.Curve(blenderActiveCurve)
        oCurve = curves.Curve(blenderOtherCurve)

        return CurvesIntersector(aCurve, oCurve)

    @staticmethod
    def ResetGlobals():
        global algoPOV
        algoPOV = None
        global algoDIR
        algoDIR = None

    @staticmethod
    def InitGlobals():
        CurvesIntersector.ResetGlobals()
        global algoPOV
        global algoDIR

        algo = bpy.context.scene.curvetools.IntersectCurvesAlgorithm
        if algo == 'From_View':
            regionView3D = util.GetFirstRegionView3D()
            if regionView3D is None:
                print("### ERROR: regionView3D is None. Stopping.")
                return

            viewPerspective = regionView3D.view_perspective
            print("--", "viewPerspective:", viewPerspective)

            if viewPerspective == 'ORTHO':
                viewMatrix = regionView3D.view_matrix
                print("--", "viewMatrix:")
                print(viewMatrix)

                algoDIR = Vector((viewMatrix[2][0], viewMatrix[2][1], viewMatrix[2][2]))
                print("--", "algoDIR:", algoDIR)

            # ## TODO: doesn't work properly
            if viewPerspective == 'PERSP':
                viewMatrix = regionView3D.view_matrix
                print("--", "viewMatrix:")
                print(viewMatrix)

                algoPOV = regionView3D.view_location.copy()
                print("--", "algoPOV:", algoPOV)

                otherPOV = Vector((viewMatrix[0][3], viewMatrix[1][3], viewMatrix[2][3]))
                print("--", "otherPOV:", otherPOV)

                localPOV = Vector((0, 0, 0))
                globalPOV = viewMatrix * localPOV
                print("--", "globalPOV:", globalPOV)

                perspMatrix = regionView3D.perspective_matrix
                print("--", "perspMatrix:")
                print(perspMatrix)

                globalPOVPersp = perspMatrix * localPOV
                print("--", "globalPOVPersp:", globalPOVPersp)

            if viewPerspective == 'CAMERA':
                camera = bpy.context.scene.camera
                if camera is None:
                    print("### ERROR: camera is None. Stopping.")
                    return

                print("--", "camera:", camera)
                cameraData = camera.data
                print("--", "cameraData.type:", cameraData.type)

                cameraMatrix = camera.matrix_world
                print("--", "cameraMatrix:")
                print(cameraMatrix)

                if cameraData.type == 'ORTHO':
                    cameraMatrix = camera.matrix_world
                    # algoDIR = Vector((cameraMatrix[2][0], cameraMatrix[2][1], cameraMatrix[2][2]))
                    algoDIR = Vector((- cameraMatrix[0][2], - cameraMatrix[1][2], - cameraMatrix[2][2]))
                    print("--", "algoDIR:", algoDIR)

                if cameraData.type == 'PERSP':
                    algoPOV = camera.location.copy()
                    print("--", "algoPOV:", algoPOV)

    def __init__(self, activeCurve, otherCurve):
        self.activeCurve = activeCurve
        self.otherCurve = otherCurve

        CurvesIntersector.InitGlobals()

    def CalcIntersections(self):
        rvIntersections1 = []
        rvIntersections2 = []

        worldMatrix1 = self.activeCurve.curve.matrix_world
        worldMatrix2 = self.otherCurve.curve.matrix_world

        for spline1 in self.activeCurve.splines:
            for spline2 in self.otherCurve.splines:
                splineIntersector = BezierSplinesIntersector(spline1, spline2, worldMatrix1, worldMatrix2)
                splineIntersections = splineIntersector.CalcIntersections()
                if splineIntersections is None:
                    continue

                spline1Intersections = splineIntersections[0]
                for splineIntersection in spline1Intersections:
                    rvIntersections1.append(splineIntersection)

                spline2Intersections = splineIntersections[1]
                for splineIntersection in spline2Intersections:
                    rvIntersections2.append(splineIntersection)

        return [rvIntersections1, rvIntersections2]

    def CalcAndApplyIntersections(self):
        mode = bpy.context.scene.curvetools.IntersectCurvesMode

        if mode == 'Empty':
            return self.CalcAndApplyEmptyAtIntersections()
        if mode == 'Insert':
            return self.CalcAndApplyInsertAtIntersections()
        if mode == 'Split':
            return self.CalcAndApplySplitAtIntersections()

        return [0, 0]

    def CalcAndApplyEmptyAtIntersections(self):
        intersections = self.CalcIntersections()
        intersectionsActive = intersections[0]
        intersectionsOther = intersections[1]

        nrActive = 0
        nrOther = 0

        affect = bpy.context.scene.curvetools.IntersectCurvesAffect

        if (affect == 'Both') or (affect == 'Active'):
            for splineIntersection in intersectionsActive:
                iPoint = splineIntersection.bezierSegmentIntersectionPoint.intersectionPoint
                bpy.ops.object.empty_add(type='PLAIN_AXES',
                                         align='WORLD',
                                         location=(iPoint.x, iPoint.y, iPoint.z), rotation=(0, 0, 0))
                nrActive += 1

        if (affect == 'Both') or (affect == 'Other'):
            for splineIntersection in intersectionsOther:
                iPoint = splineIntersection.bezierSegmentIntersectionPoint.intersectionPoint
                bpy.ops.object.empty_add(type='PLAIN_AXES',
                                         align='WORLD',
                                         location=(iPoint.x, iPoint.y, iPoint.z), rotation=(0, 0, 0))
                nrOther += 1

        return [nrActive, nrOther]

    def CalcAndApplyInsertAtIntersections(self):
        nrActive = 0
        nrOther = 0

        affect = bpy.context.scene.curvetools.IntersectCurvesAffect
        affectA = (affect == 'Both') or (affect == 'Active')
        affectO = (affect == 'Both') or (affect == 'Other')

        for iSplineA in range(len(self.activeCurve.splines)):
            splineA = self.activeCurve.splines[iSplineA]
            nrSegmentsA = len(splineA.segments)
            resPerSegA = splineA.resolutionPerSegment

            for iSplineO in range(len(self.otherCurve.splines)):
                splineO = self.otherCurve.splines[iSplineO]
                nrSegmentsO = len(splineO.segments)
                resPerSegO = splineO.resolutionPerSegment

                iSegA = 0
                while True:
                    segA = splineA.segments[iSegA]

                    iSegO = 0
                    while True:
                        segO = splineO.segments[iSegO]

                        segIntersector = BezierSegmentsIntersector(segA, segO,
                                                                   self.activeCurve.worldMatrix,
                                                                   self.otherCurve.worldMatrix)
                        segFirstIntersection = segIntersector.CalcFirstIntersection(resPerSegA, resPerSegO)

                        if segFirstIntersection is not None:
                            intPointA = segFirstIntersection[0]
                            intPointO = segFirstIntersection[1]
                            # else does something weird if 1 of them is None..
                            if (intPointA is not None) and (intPointO is not None):
                                if affectA:
                                    if intPointA is not None:
                                        splineA.InsertPoint(segA, intPointA.parameter)

                                        nrActive += 1
                                        nrSegmentsA += 1

                                if affectO:
                                    if intPointO is not None:
                                        splineO.InsertPoint(segO, intPointO.parameter)

                                        nrOther += 1
                                        nrSegmentsO += 1

                        iSegO += 1
                        if not (iSegO < nrSegmentsO):
                            break

                    iSegA += 1
                    if not (iSegA < nrSegmentsA):
                        break

                if affectO:
                    splineO.RefreshInScene()

            if affectA:
                splineA.RefreshInScene()

        return [nrActive, nrOther]

    def CalcAndApplySplitAtIntersections(self):
        nrActive = 0
        nrOther = 0

        affect = bpy.context.scene.curvetools.IntersectCurvesAffect
        affectA = (affect == 'Both') or (affect == 'Active')
        affectO = (affect == 'Both') or (affect == 'Other')

        nrSplinesA = len(self.activeCurve.splines)
        nrSplinesO = len(self.otherCurve.splines)

        iSplineA = 0
        while True:
            splineA = self.activeCurve.splines[iSplineA]
            nrSegmentsA = len(splineA.segments)
            resPerSegA = splineA.resolutionPerSegment

            iSplineO = 0
            while True:
                splineO = self.otherCurve.splines[iSplineO]
                nrSegmentsO = len(splineO.segments)
                resPerSegO = splineO.resolutionPerSegment

                iSegA = 0
                while True:
                    segA = splineA.segments[iSegA]

                    iSegO = 0
                    while True:
                        segO = splineO.segments[iSegO]

                        segIntersector = BezierSegmentsIntersector(segA, segO,
                                                                   self.activeCurve.worldMatrix,
                                                                   self.otherCurve.worldMatrix)
                        segFirstIntersection = segIntersector.CalcFirstIntersection(resPerSegA, resPerSegO)

                        if segFirstIntersection is not None:
                            intPointA = segFirstIntersection[0]
                            intPointO = segFirstIntersection[1]
                            # else does something weird if 1 of them is None..
                            if (intPointA is not None) and (intPointO is not None):
                                if affectA:
                                    if intPointA is not None:
                                        print("--", "splineA.Split():")
                                        newSplinesA = splineA.Split(segA, intPointA.parameter)
                                        if newSplinesA is not None:
                                            newResolutions = splineA.CalcDivideResolution(segA, intPointA.parameter)
                                            newSplinesA[0].resolution = newResolutions[0]
                                            newSplinesA[1].resolution = newResolutions[1]

                                            splineA = newSplinesA[0]
                                            self.activeCurve.splines[iSplineA] = splineA
                                            self.activeCurve.splines.insert(iSplineA + 1, newSplinesA[1])

                                            nrActive += 1

                                if affectO:
                                    if intPointO is not None:
                                        print("--", "splineO.Split():")
                                        newSplinesO = splineO.Split(segO, intPointO.parameter)
                                        if newSplinesO is not None:
                                            newResolutions = splineO.CalcDivideResolution(segO, intPointO.parameter)
                                            newSplinesO[0].resolution = newResolutions[0]
                                            newSplinesO[1].resolution = newResolutions[1]

                                            splineO = newSplinesO[0]
                                            self.otherCurve.splines[iSplineO] = splineO
                                            self.otherCurve.splines.insert(iSplineO + 1, newSplinesO[1])

                                            nrOther += 1

                        nrSegmentsO = len(splineO.segments)
                        iSegO += 1
                        if not (iSegO < nrSegmentsO):
                            break

                    nrSegmentsA = len(splineA.segments)
                    iSegA += 1
                    if not (iSegA < nrSegmentsA):
                        break

                nrSplinesO = len(self.otherCurve.splines)
                iSplineO += 1
                if not (iSplineO < nrSplinesO):
                    break

            nrSplinesA = len(self.activeCurve.splines)
            iSplineA += 1
            if not (iSplineA < nrSplinesA):
                break

        if affectA:
            print("")
            print("--", "self.activeCurve.RebuildInScene():")
            self.activeCurve.RebuildInScene()
        if affectO:
            print("")
            print("--", "self.otherCurve.RebuildInScene():")
            self.otherCurve.RebuildInScene()

        return [nrActive, nrOther]
